/* Modal dialog */
class ModalDialog {

    constructor() {
        this.modal_el = document.createElement('div');
        this.modal_el.id = 'modal-dialog';
        this.modal_el.setAttribute('tabindex', '-1');
        this.modal_el.setAttribute('role', 'dialog');
        this.modal_el.setAttribute('aria-modal', 'true');
        this.modal_el.setAttribute('aria-labelledby', 'modal-dialog-heading');
        this.modal_el.hidden = true;

        this.modal_heading = document.createElement('h1');
        this.modal_heading.id = 'modal-dialog-heading';
        this.modal_el.append(this.modal_heading);

        this.modal_content = document.createElement('div');
        this.modal_content.id = 'modal-dialog-content';
        this.modal_el.append(this.modal_content);

        this.modal_buttons = document.createElement('div');
        this.modal_buttons.id = 'modal-dialog-buttons';
        this.modal_el.append(this.modal_buttons);

        document.body.prepend(this.modal_el);

        this.modal_buttons.addEventListener('click', event => this._handleButtonClick(event));

        this.modal_el.addEventListener('keydown', event => this._handleModalKeydown(event));

        this.buttons = [];
        this.modal_options = {};

        this._handleBodyClick = this._handleBodyClick.bind(this);
        this._handleBodyKeydown = this._handleBodyKeydown.bind(this);
    }

    show(heading, message, buttons, options) {

        this.buttons = [];
        this.modal_options = options;
        this.modal_buttons.innerHTML = '';
        this.modal_heading.innerHTML = heading;
        this.modal_content.innerHTML = message;

        for (let i = 0; i < buttons.length; i++) {
            const button_el = document.createElement('button');
            button_el.dataset.num = i;
            button_el.innerHTML = buttons[i].text;
            buttons[i].el = button_el;
            this.buttons.push(buttons[i]);
            this.modal_buttons.append(button_el);
        }

        this.modal_el.removeAttribute('hidden');
        this._setInert([document.getElementById('manuscript')], true);
        document.body.classList.add('modal-dialog-open');

        /*
        document.body.addEventListener('click', this._handleBodyClick);
        document.body.addEventListener('keydown', this._handleBodyKeydown);
        */

        if (this._viewingInPX()) {
            this._positionForPX();
        }

        this.modal_el.focus();
    }

    close(options) {

        /*
        document.body.removeEventListener('click', this._handleBodyClick);
        document.body.removeEventListener('keydown', this._handleBodyKeydown);
        */
        
        this._setInert([document.getElementById('manuscript')], false);
        this.modal_el.setAttribute('hidden', 'true');
        document.body.classList.remove('modal-dialog-open');

        if (options && options.from) {
            options.from.focus();
        }
    }

    _handleButtonClick(event) {
        console.log('_handleButtonClick: ', event);
        const target = event.target;
        if (target.tagName.toUpperCase() !== 'BUTTON') {
            return;
        }
        const button_num = target.dataset.num;
        this.buttons[button_num].click();
    }

    _handleModalKeydown(event) {
        const target = event.target;
        const keycode = event.which;
     
        // Escape key
        if (keycode === 27) {
            event.preventDefault();
            event.stopPropagation();
            if (typeof this.modal_options.onEsc === 'function') {
                this.modal_options.onEsc();
            }
            else {
                this.close();
            }
            return;
        }

        // we only care about the Tab key
        if (keycode !== 9) {
            return;
        }

        // focus is initially on dialog div, shift+Tab: set focus to last button
        if (target.tagName.toUpperCase() !== 'BUTTON' && event.shiftKey) {
            event.preventDefault();
            event.stopPropagation();
            this.buttons[this.buttons.length - 1].el.focus();
        }
        else {
            if (event.shiftKey && target.dataset.num === '0') {
                event.preventDefault();
                this.buttons[this.buttons.length - 1].el.focus();
                event.stopPropagation();
            }
            else if (!event.shiftKey && Number(target.dataset.num) === this.buttons.length - 1) {
                event.preventDefault();
                event.stopPropagation();
                this.buttons[0].el.focus();
            }
        }

    }

    _handleBodyClick(event) {

        console.log('_handleBodyClick: ', event);

        if (!this.modal_el.contains(event.target)) {
            this.close();
        }
    }

    _handleBodyKeydown(event) {
        const keycode = e.which;
        // Escape
        if (keycode === 27) {
            this.close();
        }
    }

    _setInert(elements, set) {
        for (let el of elements) {
            console.log('_setIntert: ', el, set);
            if (set === true) {
                el.setAttribute('inert', 'true');
                el.setAttribute('hidden', 'true');
            }
            else {
                el.removeAttribute('inert');
                el.removeAttribute('hidden');
            }
        }
    }

    _viewingInPX() {
        if (typeof window.PX !== 'undefined') {
            return true;
        }
        if (typeof window.parent.PX !== 'undefined') {
            return true;
        }
        return false;
    }

    _positionForPX() {
        let fneScrollTop = undefined;
        const activeElPos = document.activeElement.getBoundingClientRect();

        try {
            const fne_content = window.parent.document.getElementById('fne-content');
            fneScrollTop = fne_content.scrollTop;
        }
        catch (e) {
                console.log('caught error trying to get fne-content: ' + e.message);
            fneScrollTop = undefined;
        }

        if (fneScrollTop !== undefined) {
            let popUpTop;

            console.log('fneScrollTop = ' + fneScrollTop + ', activeElPos.top = ' + activeElPos.top);

            if (fneScrollTop < activeElPos.top) {
                popUpTop = fneScrollTop + ((activeElPos.top - fneScrollTop) / 2)
            }
            else {
                if (activeElPos.top === 0 && activeElPos.top === fneScrollTop) {
                    try {
                        popUpTop = 100;
                    }
                    catch (e) {
                        popUpTop = fneScrollTop;
                    }
                }
                else {
                    popUpTop = fneScrollTop;
                }
            }
            this.modal_el.style.transform = 'translate(-50%)';
            this.modal_el.style.top = popUpTop + 'px';
        }

    }
}