/************************************************************
 * Standard_Dialog
 * Based on jQueryUi  ui-dialog
 *
 * How to use?
 *
 * Example:
 * Standard_Dialog.open('<div>content</div>',{width:768});
 *
 * Available options:
 * width: number
 * modal: boolean
 * from: jQueryElement,
 * draggable: boolean,
 * resizable: boolean,
 * dialogClass: string
 * buttons: [{text:'btnName', click: clickHandler}] *
 *
 **************************************************************/


var Standard_Dialog = function() {
  var DEFAULT_DIALOG_OPTIONS = {
    width: 450,
    modal: true,
    from: null,
    draggable: false,
    resizable: false,
    dialogClass: 'noclose',
    title: '',
    position: 'center',
    buttons: [
      {
        text: "OK",
        click: function() {
          Standard_Dialog.close();
        }
      }
    ]
  };

  var dialogOptions = {};

  return {
    open: function(html, options) {

      safe_log("open dialog");
      safe_log(options);

      //Standard_Dialog.close();

      options = options || {};

      if (options.buttons) {
        //options.buttons = 'OK' | ['OK','close']
        if (!Array.isArray(options.buttons)) {
          options.buttons = [].concat(options.buttons);
        }

        options.buttons = DEFAULT_DIALOG_OPTIONS.buttons.filter(function(btn) {
          return options.buttons.indexOf(btn.text) !== -1;
        });
      }

      dialogOptions = {
        width: options.width || DEFAULT_DIALOG_OPTIONS.width,
        modal: options.modal !== false,
        from: options.from || dialogOptions.from || DEFAULT_DIALOG_OPTIONS.from,
        draggable: !!options.draggable,
        resizable: !!options.resizable,
        dialogClass: options.dialogClass || DEFAULT_DIALOG_OPTIONS.dialogClass,
        title: options.title || DEFAULT_DIALOG_OPTIONS.title,
        position: options.position || DEFAULT_DIALOG_OPTIONS.position,
        buttons: options.buttons || DEFAULT_DIALOG_OPTIONS.buttons
      };

      var $uiDialogContent = $("#standard_dialog_div");
      $uiDialogContent.dialog("close");
      $uiDialogContent.remove();

      var labelledElem = 'ui-dialog-title-standard_dialog_div ';
      if (dialogOptions.from) {
        if (dialogOptions.from[0] === $('#menu-button')[0]) {
          labelledElem += 'toc_head '
        }
      }
      labelledElem += 'standard_dialog_div';

      $uiDialogContent = $("<div>")
        .attr('id', 'standard_dialog_div')
        .html(html);

      if (dialogOptions.modal) {
        $('body').css('overflow', 'hidden');
      }
      $("body").append($uiDialogContent);

      $uiDialogContent.dialog(dialogOptions);

      var $uiDialog = $('.ui-dialog');

      //remove close buttons from dialog title
      if (dialogOptions.dialogClass.indexOf('noclose') !== -1) {
        $uiDialog.find('.ui-dialog-titlebar-close').remove();
      }

      if (isIE) {
        // Jaws has limited length for aria-label
        // so to read the whole text in the window we need to add it to aria-live
        player.setLiveText($('#ui-dialog-title-standard_dialog_div').text());
        if (dialogOptions.from && dialogOptions.from[0] === $('#menu-button')[0]) {
          player.setLiveText($('#toc_head').text())
        }
        player.setLiveText($('#standard_dialog_div').text())
      } else {
        $uiDialog.attr('aria-labelledby', labelledElem); //jaws18, nvda, ios VO works
      }

      if (isAndroid) {
        $uiDialog.focus();
      } //android needs to focus on element

      if (document.all) {
        $uiDialog.css('top', '255px');
      }

      $uiDialog.trapTabKey();

      $('body > *:not(.ui-dialog, link, script, br, .visually-hidden)').attr('aria-hidden', true);

      $(document).bind('keydown', function(e) {
        if (e.keyCode === 27) {
          Standard_Dialog.close();
        }
      })
    },

    // for now alert is just like any other dialog; we could change that later.
    wait: function(html, options) {
      this.open(html, options);
    },

    alert: function(html, options) {
      safe_log("alert dialog");
      this.open(html, options);
    },

    close: function() {
      safe_log("close dialog");
      safe_log(dialogOptions);

      // when we close from here, we *don't* call the .dialog("close") fn, so if you need to call the fn defined
      // as the onclose event when you called Standard_Dialog.open(), you should do that separately
      // $("#standard_dialog_div").dialog("close");
      $("#standard_dialog_div").remove();
      if (dialogOptions.from) {
        safe_log("custom focus");
        safe_log(dialogOptions);
        dialogOptions.from.focus();
      }
      else {
        safe_log("setting focus to menu toc button");
        $('#menu_button').focus();
      }

      $('body > *:not(.ui-dialog, link, script, br, .visually-hidden)').attr('aria-hidden', false);

      if (dialogOptions.modal) {
        $('body').css('overflow', 'auto');
      }

      $(document).unbind('keydown');
    }
  };
}();
