/* reflection_activity.js */

(function($) {
    $.fn.changeElementType = function(newType) {
        var attrs = {};

        $.each(this[0].attributes, function(idx, attr) {
            attrs[attr.nodeName] = attr.nodeValue;
        });

        this.replaceWith(function() {
            return $("<" + newType + "/>", attrs).append($(this).contents());
        });
    };
})(jQuery);

// Fix HTML validation errors

$('[data-type="query"]').each(function () {
    $(this).changeElementType('div');
});

$('[data-type="section"]').removeAttr('level');

$('[http-equiv="X-UA-Compatible"]').remove();

$('img[data-figure-id]').each(function () {
    var $this = $(this);
    var id = $this.attr('data-figure-id');
    $this.attr('id', 'img' + id);
});

$('[data-type="field"]').each(function () {
    var $this = $(this);
    var key = $this.attr('key');
    $this.removeAttr('key');
    $this.attr('data-key', key);
});

var CB_QUERY_TEXT = {
    '0': "Did Goal Setting improve?",
    '1': "Did Academic and Financial Planning improve?",
    '2': "Did Motivation, Decision Making, and Personal Responsibility improve?",
    '3': "Did Organization and Time Management improve?",
    '4': "Did Learning/Critical Thinking improve?",
    '5': "Did Reading improve?",
    '6': "Did Note Taking improve?",
    '7': "Did Memory and Studying improve?",
    '8': "Did Test Taking improve?",
    '9': "Did Information Literacy and Communication improve?",
    '10': "Did Social Belonging improve?",
    '11': "Did Health/Wellness improve?",
    '12': "Did Career Readiness improve?",
    '13': "Did Online Learning improve?",


    '14': "Did Goal Setting go down?",
    '15': "Did Academic and Financial Planning go down?",
    '16': "Did Motivation, Decision Making, and Personal Responsibility go down?",
    '17': "Did Organization and Time Management go down?",
    '18': "Did Learning/Critical Thinking go down?",
    '19': "Did Reading go down?",
    '20': "Did Note Taking go down?",
    '21': "Did Memory and Studying go down?",
    '22': "Did Test Taking go down?",
    '23': "Did Information Literacy and Communication go down?",
    '24': "Did Social Belonging go down?",
    '25': "Did Health/Wellness go down?",
    '26': "Did Career Readiness go down?",
    '27': "Did Online Learning go down?",

    
    '28': "Did Goal Setting stay the same?",
    '29': "Did Academic and Financial Planning stay the same?",
    '30': "Did Motivation, Decision Making, and Personal Responsibility stay the same?",
    '31': "Did Organization and Time Management stay the same?",
    '32': "Did Learning/Critical Thinking stay the same?",
    '33': "Did Reading stay the same?",
    '34': "Did Note Taking stay the same?",
    '35': "Did Memory and Studying stay the same?",
    '36': "Did Test Taking stay the same?",
    '37': "Did Information Literacy and Communication stay the same?",
    '38': "Did Social Belonging stay the same?",
    '39': "Did Health/Wellness stay the same?",
    '40': "Did Career Readiness stay the same?",
    '41': "Did Online Learning stay the same?",


    '54': "Focus on Goal Setting?",
    '55': "Focus on Academic and Financial Planning?",
    '56': "Focus on Motivation, Decision Making, and Personal Responsibility?",
    '57': "Focus on Organization and Time Management?",
    '58': "Focus on Learning/Critical Thinking?",
    '59': "Focus on Reading?",
    '60': "Focus on Note Taking?",
    '61': "Focus on Memory and Studying?",
    '62': "Focus on Test Taking?",
    '63': "Focus on Information Literacy and Communication?",
    '64': "Focus on Social Belonging?",
    '65': "Focus on Health/Wellness?",
    '66': "Focus on Career Readiness?",
    '67': "Focus on Online Learning?",

};


// Detect platform with https://github.com/bestiejs/platform.js
var iOS = platform.os.family === 'iOS';           //detect iOS (VoiceOver)
var isMac = platform.os.family === 'OS X';         //detect MacOS (VoiceOver)
var isAndroid = platform.os.family === 'Android'; //detect Android (TalkBack)
var isFirefox = platform.name === 'Firefox';      //detect Firefox (NVDA)
var isIE = platform.name === 'IE';                //detect IE (Jaws)

if (isIE) {
    //prevents focusing on the body in the IE
    document.body.setAttribute('tabindex', '-1');
}

var noopApi = {
    initializeClass: function() {
        var dfd = jQuery.Deferred();
        setTimeout(function() {
            dfd.resolve(null);
        }, Math.floor(100 + Math.random() * 500));
        return dfd.promise();
    },
    initializeUser: function() {
        var dfd = jQuery.Deferred();
        setTimeout(function() {
            dfd.resolve(null);
        }, Math.floor(100 + Math.random() * 500));
        return dfd.promise();
    },
    loadClassData: function() {
        var dfd = jQuery.Deferred();
        setTimeout(function() {
            dfd.resolve([]);
        }, Math.floor(100 + Math.random() * 500));
        return dfd.promise();
    },
    loadGlobalClassData: function() {
        var dfd = jQuery.Deferred();
        setTimeout(function() {
            dfd.resolve(null);
        }, Math.floor(100 + Math.random() * 500));
        return dfd.promise();
    },
    setData: function() {},
    setGlobalData: function() {
        var dfd = jQuery.Deferred();
        setTimeout(function() {
            dfd.resolve(null);
        }, Math.floor(100 + Math.random() * 500));
        return dfd.promise();
    }
};

var Activity_reflection = Activity_manuscript_type.extend({

    question_submit_button: function (q) {

        var button_html = "<button class='standard_button question_submit_button' type='button' id='question_submit_button_" + q.index + "'>Submit<span class='visually-hidden'>, question " + q.index + "</span></button>" ;

        return button_html;

    },


    submit_question: function (question_index) {

        player.setLiveText('');


        if (typeof question_index == "object") {
            // extract the question_index from the object
            question_index = question_index.data.question_index;
        }

        $('#question_error_' + question_index).text('');

        var query = player.activity.questions[question_index].queries[0];

        // clear any essay question feedback
        $('#essay_question_evaluation_feedback_' + (question_index - 1)).text('');

        if (query.getQueryTypeForARGA() === 'essay') {
            var userAnswer = $('#query_answer_' + query.query_index).val().trim();
            if (!(/\S/.test(userAnswer))) {
                //query.set_evaluation_feedback("You must enter an answer before submitting.");
                $('#question_error_' + question_index).text('You must enter an answer before submitting.');
                return;
            }
        }
        /*
        else if (player.activity.questions[question_index].queries[0].getQueryTypeForARGA() === 'multiple choice') {
            if (!player.activity.questions[question_index].queries[0].userHasSelectedAnswer()) {
                player.activity.questions[question_index].queries[0].set_evaluation_feedback("You must select an answer.");
                return;
            }
        }
        */

        var buttons = [
            {
                text: "No",
                click: function () {
                    player.modal_dialog.close({ from: $('#question_submit_button_' + question_index) });
                    //Standard_Dialog.close();
                }
            },
            {
                text: "Yes",
                click: function () {
                    player.modal_dialog.close();
                    // Standard_Dialog.close();
                    player.activity.submit_question_go(question_index);
                }
            }
        ];

        if (query.getQueryTypeForARGA() === 'essay') {

            player.modal_dialog.show('Confirm Submission',
                'Do you want to submit your answer? You cannot edit your answer after submitting.',
                buttons,
                {
                    onEsc: function () {
                        player.modal_dialog.close({ from: $('#question_submit_button_' + question_index) })
                    }
                }
            );
            /*
            Standard_Dialog.open('Do you want to submit your answer? You cannot edit your answer after submitting.',
                {
                    buttons: buttons,
                    onEsc: function () {
                        Standard_Dialog.close();
                        $('#query_answer_' + question_index).focus();
                    },
                    diagTitle: 'Confirm submission',
                    from: $('#question_submit_button_' + question_index)
                });
            */
        }
        else {
            player.activity.submit_question_go(question_index);
        }
    },

    submit_question_go: function(question_index) {

        /* begin original submit_question() */

        // the arg may be a number or an event object; in the latter case...
        if (typeof question_index == "object") {
            // extract the question_index from the object
            question_index = question_index.data.question_index;
        }

        var t = player.activity;

        var question = t.questions[question_index];

        var all_answers_correct = true;
        // for each query in the question...
        for (var i = 0; i < question.queries.length; ++i) {
            var query = question.queries[i];

            // evaluate the answer. If evaluateAnswer returns false, the answer isn't submittable at this time -- 
            // most likely because it's a free-response answer that needs to be evaluated via MAVE.
            var answer_submittable = query.evaluateAnswer(question_index);
            if (answer_submittable === false) {
                // as soon as we get an unsubmittable answer, we return. submit_question
                // will be called again, either by the callback fn of MAVE or because the student
                // resubmits. At that point we'll catch any additional free-response answers that need
                // checking, or we'll finish this loop here and move on.
                return;
            }
            all_answers_correct = all_answers_correct && query.isCorrect();
        }

        // if we made it to here, we're at least going to show which items are correct and incorrect
        // we might overwrite with review mode below, but that's OK
        for (var i = 0; i < question.queries.length; ++i) {
            var query = question.queries[i];
            query.jq.html(query.getHTML("review_correct_incorrect"));
        }

        // update question tries
        ++question.tries;

        // if answer is incorrect, and max_tries is set, and # of tries is less than max_tries
        // (and the user didn't just give up)
        if ((!all_answers_correct && question.md.max_tries != null && question.tries < question.md.max_tries) && question.gave_up != true) {
            // tell user he can try again
            t.show_question_feedback(question, false, true);
            return;
        } else {
            t.show_question_feedback(question, all_answers_correct, false);
        }

        // if we make it to here we need to set ARGA info for the question, and show queries in review mode
        // if md.question_report_text is set, we submit the question as a whole, rather than as individual queries
        var query_correct_answers = "";
        var query_responses = "";
        var query_responses_for_questionData = "";
        var query_grade_sum = 0;
        // for each query in the question...
        for (var i = 0; i < question.queries.length; ++i) {
            var query = question.queries[i];

            // establish a question number to use
            var query_number = t.get_query_number(question, i);

            // get a query weight
            var query_weight = question.get_weight();
            // if the question has more than one query, divide weight accordingly
            if (question.queries.length > 1) {
                query_weight = Math.round(query_weight / question.queries.length);
            }

            // if credit_after_max_tries is set, give at least that credit now, as long as the student tried
            if (question.md.credit_after_max_tries != null) {
                var query_grade = query.getGradeForARGA();
                if (query_grade >= 0 && query_grade < question.md.credit_after_max_tries) {
                    query.overrideGrade(question.md.credit_after_max_tries * 1);
                }
            }

            // set the ARGA response, unless we're saving one response for the whole question
            if (!question.md_set("question_report_text")) {
                t.ARGA_set_query_response(query, query_number, query_weight);

                // otherwise compose query roll-ups
            } else {
                if (query_correct_answers != "") {
                    query_correct_answers += " / ";
                    query_responses += " / ";
                    query_responses_for_questionData += t.query_responses_separator;
                }
                query_correct_answers += query.getCorrectAnswerForARGA();
                query_responses += query.getUserAnswerForARGA();
                // see restore_saved_answers for how this is used later
                query_responses_for_questionData += query.getUserAnswerForARGA()
                    + t.query_responses_grade_separator
                    + query.getOriginalGrade();
                query_grade_sum += query.getGradeForARGA() * 1;
            }

            // show in review mode, unless question metadata says not to show correct answers
            if (question.md.show_correct_answers != "false") {
                query.jq.html(query.getHTML("review"));
            }
        }

        // if we're saving one response for the whole question, save here
        if (question.md_set("question_report_text")) {
            var question_grade = Math.round(query_grade_sum / question.queries.length);
            t.ARGA_set_question_response(question, query_correct_answers, query_responses, question_grade, query_responses_for_questionData);
        }

        question.jq.addClass('answered');

        // grade the activity
        t.grade_activity();

        // then if ARGA is running, submit data; show_grade_saved_message will be called on completion.
        if (player.ARGA_running) {
            t.ARGA_submit_data();
        }

        // update the question_action_div
        t.update_question_action_div(question_index);

        // if the item is in a sequence, show the next question
        t.show_next_question(question);

        // update section status
        player.update_section_status();

        /* end original submit_question() */

        // the arg may be a number or an event object; in the latter case...
        if (typeof question_index == "object") {
            // extract the question_index from the object
            question_index = question_index.data.question_index;
        }

        // if question has been answered
        var question = player.activity.questions[question_index];
        if (question.is_complete()) {
            // remove submit button

           $('#question_submit_button_' + question_index).remove();

            // set focus on next button if all questions on page have been answered
            // 20210407: Not doing this any more, setting focus to feedback
            /*
            var curr_sec = player.get_current_section();
            if (curr_sec.all_questions_answered) {
                $('#next_button').focus();
            }
            */
            
            var curr_sec = player.get_current_section();
            console.log('curr_sec: ', curr_sec);
            if (curr_sec.all_questions_answered) {
                setTimeout(function () {
                    console.log('setting focus on next button');
                    $('#next_button').focus();
                }, 250);
                //$('#question_feedback_' + question_index).focus();
            }
            else {
                console.log('setting focus on question feedback');
                $('#question_feedback_' + question_index).focus();
            }
        }

    },


    initialize: function() {

        this._super();

        player.fix_local_links();

        //console.log('finding all question_action_divs');

        $('.question_action_div').each(function () {
            var $this = $(this);
            //console.log('found question_action_div: ', $this);
            var id = $this.attr('id');
            var fb_id = id.replace(/action_div/, 'feedback');
            var err_id = id.replace(/action_div/, 'error');
            $this.prepend('<div class="question_error" role="alert" tabindex="-1" id="' + err_id + '"></div><div class="question_feedback" role="status" tabindex="-1" id="' + fb_id + '"></div>');
        });
    
    },


    show_question_feedback: function(question, is_correct, try_again) {
	// remove prior feedback if there
	//$("#question_feedback_" + question.index).remove();

	// if we don't have any feedback at all, return
	if (question.md.feedback == null && question.md.feedback_correct == null && question.md.feedback_incorrect == null && question.md.feedback_hint == null) {
	    return;
	}

        var fb_str = "";

	var html = "<div class='question_feedback' tabindex='-1' id='question_feedback_" + question.index + "'>";
	//+ "<div class='question_feedback_head'>" + this.feedbackTitle + "</div>";
    if (question.md.feedback_header === null || question.md.feedback_header === undefined ||
        question.md.feedback_header !== "false") {
        html += "<h3 class='visually-hidden'>Question Feedback</h3>";
    }

	// include general feedback if there
	if (question.md.feedback != null) {
        html += "<div class='question_feedback_part'>" + question.md.feedback + "</div>";
        fb_str += question.md.feedback;
	}

	// include correct feedback if there and the answer is correct
	if (is_correct && question.md.feedback_correct != null) {
	    html += "<div class='question_feedback_part'>" + question.md.feedback_correct + "</div>";
	}

	// include incorrect or hint feedback if there and the answer is incorrect
	if (!is_correct) {
	    var found_feedback_string = false;
	    // if user gets to try again...
	    if (try_again) {
		// check for feedback_conditional's
		if (question.md.feedback_conditional.length > 0) {
		    for (var i in question.md.feedback_conditional) {
			var fbc = question.check_conditional(question.md.feedback_conditional[i]);
			if (fbc != "") {
			    html += "<div class='question_feedback_part'>" + fbc + "</div>";
			    found_feedback_string = true;
			    break;
			}
		    }
		}

		// if we didn't find a feedback string to use above, check for feedback_hint
		if (!found_feedback_string && question.md.feedback_hint != null) {
		    html += "<div class='question_feedback_part'>" + question.md.feedback_hint + "</div>";
		    found_feedback_string = true;
		}
	    }

	    // if we didn't find a feedback string to use above, check for feedback_incorrect
	    if (!found_feedback_string && question.md.feedback_incorrect != null) {
		html += "<div class='question_feedback_part'>" + question.md.feedback_incorrect + "</div>";
	    }
	}

	// close question_feedback div
	html += "</div>";

	// append the div, then fade it in
	//$("#question_action_div_" + question.index).prepend(html);
	//$("#question_feedback_" + question.index).fadeIn("fast", function() { });

        console.log('setting feedback status');
        $("#question_feedback_" + question.index).text(fb_str);

        // announce feedback for screen readers

        // 20210407: We are setting focus on feedback instead.
        /*
        if (question.md.feedback !== null) {
            player.setLiveText(question.md.feedback);
        }
        */

        // if this is question 1 then we want to change the <label>s
        // to spans
        /*
        if (question.index == 1) {
            $('#question_1_table label').each(function() {
                $(this).changeElementType('span');
            });
        }
        */

    },

    update_question_action_div: function(question_index) {
	var question = this.questions[question_index];

	// calculate points_awarded and points_possible
	var points_awarded = 0;
	var points_possible = 0;
	var points_provisional = false;
	for (var i = 0; i < question.queries.length; ++i) {
	    var query = question.queries[i];
	    points_awarded += query.getPointsAwarded();
	    points_possible += query.getPointsPossible();
	    if (query.pointsAreProvisional() == true) {
		points_provisional = true;
	    }
	}

	// round to at most two decimal places
	points_awarded = Math.round(points_awarded * 100) / 100;
	points_possible = Math.round(points_possible * 100) / 100;

	// hide submit button (and any other similar buttons that might be added by subtypes), but show points earned
	var jq = $("#question_action_div_" + question.index);
	jq.find(".question_submit_button").hide();
	var points_html = "<div class='points_awarded_div'>Points awarded: " + points_awarded + " out of " + points_possible;

	if (points_provisional) {
	    points_html += " (<a href='javascript:player.activity.explain_provisional()'>provisional</a>)";
	}
	points_html += "</div>";

	var jq = $("#question_action_div_" + question.index);
	jq.find(".points_awarded_div").remove();

        // we are not showing points
        if (jq.find(".question_feedback").length > 0) {
	    //jq.find(".question_feedback").after(points_html);
	} else {
	    //jq.prepend(points_html);
	}
    }

});

var Player_reflection = Player_manuscript_type.extend({

    ACTIVITY_COMPLETED: false,

    INSTRUCTOR_NAV_MESG_SHOWN: false,

    POPUP_IS_SHOWING: false,

    INITIAL_VIEW: true,

    isInstructor: false,

    modal_dialog: undefined,

    measure_names: [{
        key: '2',
        value: 'Critical Thinking/Goal Setting',
    }, {
        key: '3',
        value: 'Motivation/Decision Making/Personal Responsibility',
    }, {
        key: '4',
        value: 'Learning Preferences',
    }, {
        key: '5',
        value: 'Organization/Time Management',
    }, {
        key: '6',
        value: 'Reading',
    }, {
        key: '7',
        value: 'Note Taking',
    }, {
        key: '8',
        value: 'Memory/Studying',
    }, {
        key: '9',
        value: 'Test Taking',
    }, {
        key: '10',
        value: 'Information Literacy and Communication',
    }, {
        key: '11',
        value: 'Connecting with Others',
    }, {
        key: '12',
        value: 'Personal and Financial Health',
    }, {
        key: '13',
        value: 'Academic and Career Planning',
    }],

    measure_names_new: [{
        key: '2',
        value: 'Goal Setting',
        score: 0
    },
    {
        key: '3',
        value: 'Academic and Financial Planning',
        score: 0
    },
    {
        key: '4',
        value: 'Motivation, Decision Making, and Personal Responsibility',
        score: 0
    },
    {
        key: '5',
        value: 'Organization and Time Management',
        score: 0
    },
    {
        key: '6',
        value: 'Learning/Critical Thinking',
        score: 0
    },
    {
        key: '7',
        value: 'Reading',
        score: 0
    },
    {
        key: '8',
        value: 'Note Taking',
        score: 0
    },
    {
        key: '9',
        value: 'Memory and Studying',
        score: 0
    },
    {
        key: '10',
        value: 'Test Taking',
        score: 0
    },
    {
        key: '11',
        value: 'Information Literacy and Communication',
        score: 0
    },
    {
        key: '12',
        value: 'Social Belonging',
        score: 0
    },
    {
        key: '13',
        value: 'Health/Wellness',
        score: 0
    },
    {
        key: '14',
        value: 'Career Readiness',
        score: 0
    },
    {
        key: '15',
        value: 'Online Learning',
        score: 0
    }],

    scale_percentages: {
        2: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 5, 7, 10, 15, 21, 29, 39, 49, 59, 69, 78, 85, 91, 94, 97, 98, 99],
        3: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4, 6, 9, 13, 17, 23, 30, 36, 44, 52, 58, 68, 76, 82, 87, 92, 96, 98, 99, 99],
        4: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 5, 7, 10, 14, 20, 26, 33, 41, 49, 57, 63, 70, 76, 82, 86, 90, 93, 95, 97, 98, 99, 99, 99, 99, 99],
        5: [1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4, 6, 8, 12, 16, 20, 25, 31, 37, 43, 48, 55, 61, 66, 72, 77, 81, 85, 88, 91, 93, 95, 96, 97, 98, 99, 99, 99, 99, 99, 99, 99],
        6: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 3, 5, 7, 10, 12, 16, 21, 28, 34, 42, 50, 58, 67, 74, 81, 87, 91, 94, 96, 97, 98, 99, 99, 99],
        7: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4, 6, 8, 11, 15, 19, 24, 30, 37, 45, 53, 61, 70, 77, 84, 88, 92, 95, 97, 98, 99, 99],
        8: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 5, 7, 11, 15, 21, 27, 36, 45, 55, 65, 73, 80, 87, 92, 95, 97, 98, 99, 99, 99],
        9: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 3, 5, 7, 9, 12, 16, 21, 27, 33, 40, 48, 56, 65, 71, 78, 85, 90, 93, 95, 97, 98, 99, 99, 99],
        10: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 5, 7, 10, 13, 17, 22, 28, 34, 41, 49, 56, 63, 70, 77, 82, 87, 91, 94, 96, 97, 99, 99, 99, 99],
        11: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 7, 10, 13, 16, 21, 27, 34, 42, 50, 58, 67, 74, 80, 86, 90, 94, 96, 98, 99, 99, 99],
        12: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 5, 7, 10, 13, 16, 21, 25, 30, 36, 43, 51, 59, 65, 71, 77, 83, 87, 91, 94, 96, 98, 99, 99, 99, 99, 99],
        13: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 3, 4, 6, 7, 9, 11, 15, 18, 22, 27, 32, 39, 45, 51, 57, 63, 69, 75, 80, 84, 88, 91, 93, 95, 97, 98, 99, 99, 99]
    },

    scale_percentages_new: {
        2: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 12, 14, 19, 24, 24, 31, 40, 55, 64, 71, 83, 88, 93, 98, 99],
        3: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 5, 7, 7, 12, 14, 19, 21, 24, 36, 43, 48, 52, 60, 64, 79, 88, 88, 93, 93, 95, 95, 95, 98, 99],
        4: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 5, 7, 9, 13, 17, 22, 28, 35, 42, 51, 59, 67, 74, 81, 87, 92, 96, 99],
        5: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4, 6, 8, 11, 14, 18, 23, 27, 32, 38, 43, 49, 54, 59, 64, 69, 73, 77, 80, 84, 87, 89, 92, 94, 95, 97, 97, 98, 99, 99, 99],
        6: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 7, 10, 10, 17, 24, 24, 31, 38, 48, 60, 67, 79, 83, 86, 90, 98, 98, 99],
        7: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4, 6, 8, 11, 14, 18, 24, 30, 37, 45, 53, 62, 70, 78, 84, 89, 92, 95, 97, 98, 99],
        8: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 7, 9, 11, 14, 18, 22, 27, 33, 40, 47, 55, 63, 70, 78, 83, 88, 91, 94, 96, 98, 99],
        9: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4, 6, 8, 12, 16, 22, 29, 38, 47, 57, 66, 75, 83, 89, 92, 95, 97, 98, 99],
        10: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 3, 5, 7, 9, 12, 16, 20, 26, 32, 38, 46, 54, 61, 69, 77, 83, 87, 90, 93, 95, 97, 99],
        11: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 5, 6, 9, 11, 15, 19, 24, 30, 36, 43, 50, 57, 64, 71, 77, 83, 87, 91, 94, 96, 97, 98, 99],
        12: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 5, 7, 21, 21, 26, 36, 43, 52, 57, 60, 64, 79, 88, 93, 93, 95, 98, 99, 99, 99],
        13: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 5, 5, 7, 12, 19, 21, 21, 21, 38, 45, 52, 64, 64, 74, 76, 81, 83, 93, 93, 95, 95, 98, 99, 99, 99, 99, 99, 99],
        14: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 5, 7, 7, 17, 17, 17, 21, 26, 31, 36, 38, 43, 48, 55, 57, 64, 69, 79, 83, 86, 90, 95, 95, 95, 98, 99],
        15: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 4, 5, 6, 7, 11, 13, 18, 22, 29, 37, 49, 61, 70, 79, 85, 89, 92, 95, 99]
    },

    clean_query_text: function(query_text) {
        query_text = query_text.replace(/<strong>blank<\/strong>/g, "");
        query_text = query_text.replace(/<span [^>]+>/g, "");
        query_text = query_text.replace(/<\/span>/g, "");
        query_text = query_text.replace(/<em>/g, "");
        query_text = query_text.replace(/<\/em>/g, "");
        query_text = query_text.replace(/<br\/>/g, "");
        query_text = query_text.replace(/<br>/g, "");
        query_text = query_text.replace(/<\/strong>/g, "");
        query_text = query_text.replace(/<strong>/g, "");
        query_text = query_text.replace(/’/g, "'");
        query_text = query_text.replace(/“/g, "\"");
        query_text = query_text.replace(/”/g, "\"");
        query_text = query_text.replace(/—/g, "--");

        // customization for this activity

        // Essay questions have a link to slide 3 which doesn't need to
        // be in the grade book
        query_text = query_text.replace(/Revisit Slide 3 .*$/, "");

        safe_log("clean_query_text: " + query_text);

        return query_text;
    },

    // DataServices object
    ds: {

        abortActivity: false,

        api: (typeof((((window.parent || {}).PX || {}).API || {}).DataService) != 'undefined') ? window.parent.PX.API.DataService : noopApi,

        generateDataRecord: function(datasetID, key, value, loadOnUserInit, loadOnClassInit, loadOnSchoolInit) {

            var r = {
                CourseId: parent.PxPage.Context.Course.EntityId,
                InstitutionId: null,
                UserId: parent.PxPage.Context.User.CurrentUserId,
                Key: key,
                Value: {
                    v: JSON.stringify(value)
                },
                DataSetId: datasetID,
                LoadOnUserInit: loadOnUserInit,
                LoadOnClassInit: loadOnClassInit,
                LoadOnSchoolInit: loadOnSchoolInit

            };
            return r;

        },

        preUserData: {},
        postUserData: {},

        initialize: function() {

            this.initializeUser(function() {

            });

        }, // end initialize

        preTestIsComplete: false,
        postTestIsComplete: false,
        instructorPreTestIsComplete: true,
        instructorPostTestIsComplete: true,

        getPreTestData: function() {

            //safe_log("getPreTestData");

            this.api.initializeUser('aces').then(
                // success
                function(a) {

                    // if we got pre-test data for user
                    if (a !== null && a !== undefined) {
                        for (var i = 0; i < a.length; i++) {
                            //safe_log(a[i].Key);
                            //safe_log(a[i].Value.v);

                            // brb: This should be in a try/catch block
                            // and then we log any errors but we don't have
                            // a way to log in PX right now
                            player.ds.preUserData[a[i].Key] = (JSON.parse(a[i].Value.v));
                        }

                        // now check that the pre-test is complete
                        var isComplete = 0;
                        try {
                            isComplete = player.ds.preUserData.scores.isComplete;
                            if (isComplete !== 1) {
                                if (player.isInstructor) {
                                    player.ds.generateTestScores('pre');
                                    player.ds.getPostTestData();
                                }
                                else {

                                    player.modal_dialog.show('Prerequisite Not Met',
                                        'You must complete the ACES Pre-Test before you can attempt this activity.',
                                        [
                                            {
                                                text: "Close",
                                                click: function () {
                                                    player.modal_dialog.close()
                                                }
                                            }
                                        ],
                                        {
                                            onEsc: function () {
                                                player.modal_dialog.close()
                                            }
                                        }
                                    );

                                    /*
                                    Standard_Dialog.open("You must complete the ACES Pre-Test before you can attempt this activity.", {
                                        title: 'Prerequisite Not Met'
                                        //from: $('#ds_fail')
                                    });
                                    */
                                    brbDisableActivity('Please exit out of this activity and complete the ACES Pre-Test.', 'Pre-Test not completed', 'isComplete = ' + isComplete);
                                }
                            }
                            // if we get here then the user has completed
                            // the pre-test and we can move on to getting
                            // post-test data
                            else {
                                player.ds.preTestIsComplete = true;
                                //safe_log("user has completed pre-test, checking post-test");
                                player.ds.getPostTestData();
                            }
                        } catch (err) {
                            if (player.isInstructor) {
                                player.ds.generateTestScores('pre');
                                player.ds.getPostTestData();
                            }
                            else {

                                player.modal_dialog.show('Prerequisite Not Met',
                                    'You must complete the ACES Pre-Test before you can attempt this activity.',
                                    [
                                        {
                                            text: "Close",
                                            click: function () {
                                                player.modal_dialog.close()
                                            }
                                        }
                                    ],
                                    {
                                        onEsc: function () {
                                            player.modal_dialog.close()
                                        }
                                    }
                                );

                                /*
                                Standard_Dialog.open("You must complete the ACES Pre-Test before you can attempt this activity.", {
                                    title: 'Prerequisite Not Met'
                                    //from: $('#ds_fail')
                                });
                                */
                                brbDisableActivity('Please exit out of this activity and complete the ACES Pre-Test.', 'Pre-Test not completed', 'Error retrieving isComplete', err);
                            }
                        }
                    }
                    // if we did NOT get pre-test data for user
                    else {
                        if (player.isInstructor) {
                            player.ds.generateTestScores('pre');
                            player.ds.getPostTestData();
                        }
                        else {

                            player.modal_dialog.show('Prerequisite Not Met',
                                'You must complete the ACES Pre-Test before you can attempt this activity.',
                                [
                                    {
                                        text: "Close",
                                        click: function () {
                                            player.modal_dialog.close()
                                        }
                                    }
                                ],
                                {
                                    onEsc: function () {
                                        player.modal_dialog.close()
                                    }
                                }
                            );

                            /*
                            Standard_Dialog.open("You must complete the ACES Pre-Test before you can attempt this activity.", {
                                title: 'Prerequisite Not Met'
                                //from: $('#ds_fail')
                            });
                            */
                            brbDisableActivity('Please exit out of this activity and complete the ACES Pre-Test.', 'Pre-Test not completed', 'No DS info found for user');
                        }
                    }
                },
                // failure
                function(a, b, c) {

                    if (typeof a === 'object') {
                        a = JSON.stringify(a);
                    } else if (a === undefined) {
                        a = '';
                    }
                    b = b === undefined ? '' : b;
                    c = c === undefined ? '' : c;

                    player.modal_dialog.show('System Error',
                        'There was an error retrieving your data. Please exit out of this activity and then open it again to resume.',
                        [
                            {
                                text: "Close",
                                click: function () {
                                    player.modal_dialog.close()
                                }
                            }
                        ],
                        {
                            onEsc: function () {
                                player.modal_dialog.close()
                            }
                        }
                    );
                    
                    /*
                    Standard_Dialog.open("There was an error retrieving your data.", {
                        title: 'System Error'
                        //from: $('#ds_fail')
                    });
                    */
                    brbDisableActivity('Please exit out of this activity and then open it again to resume.', 'initializeUser failed for pre-test', a, b, c);
                }
            );
        },


        getPostTestData: function() {

            //safe_log("getPostTestData...");

            this.api.initializeUser('acespost').then(
                // success
                function(a) {

                    // if we got post-test data for user
                    if (a !== null && a !== undefined) {
                        for (var i = 0; i < a.length; i++) {
                            //safe_log(a[i].Key);
                            //safe_log(a[i].Value.v);

                            // brb: This should be in a try/catch block
                            // and then we log any errors but we don't have
                            // a way to log in PX right now
                            player.ds.postUserData[a[i].Key] = (JSON.parse(a[i].Value.v));
                        }

                        // now check that the post-test is complete
                        var isComplete = 0;
                        try {
                            isComplete = player.ds.postUserData.scores.isComplete;
                            if (isComplete !== 1) {
                                if (player.isInstructor) {
                                    player.ds.generateTestScores('post');
                                }
                                else {

                                    player.modal_dialog.show('Prerequisite Not Met',
                                        'You must complete the ACES Post-Test before you can attempt this activity.',
                                        [
                                            {
                                                text: "Close",
                                                click: function () {
                                                    player.modal_dialog.close()
                                                }
                                            }
                                        ],
                                        {
                                            onEsc: function () {
                                                player.modal_dialog.close()
                                            }
                                        }
                                    );

                                    /*
                                    Standard_Dialog.open("You must complete the ACES Post-Test before you can attempt this activity.", {
                                        title: 'Prerequisite Not Met'
                                        //from: $('#ds_fail')
                                    });
                                    */
                                    brbDisableActivity('Please exit out of this activity and complete the ACES Post-Test.', 'Post-Test not completed', 'isComplete = ' + isComplete);
                                }
                            }
                            // if we get here then the user has completed
                            // the pre-test and we can move on to getting
                            // post-test data
                            else {
                                //safe_log("user has completed post-test");
                                player.ds.postTestIsComplete = true;
                            }
                        } catch (err) {
                            if (player.isInstructor) {
                                player.ds.generateTestScores('post');
                            }
                            else {

                                player.modal_dialog.show('Prerequisite Not Met',
                                    'You must complete the ACES Post-Test before you can attempt this activity.',
                                    [
                                        {
                                            text: "Close",
                                            click: function () {
                                                player.modal_dialog.close()
                                            }
                                        }
                                    ],
                                    {
                                        onEsc: function () {
                                            player.modal_dialog.close()
                                        }
                                    }
                                );

                                /*
                                Standard_Dialog.open("You must complete the ACES Post-Test before you can attempt this activity.", {
                                    title: 'Prerequisite Not Met'
                                    //from: $('#ds_fail')
                                });
                                */
                                brbDisableActivity('Please exit out of this activity and complete the ACES Post-Test.', 'Post-Test not completed', 'Error retrieving isComplete', err);
                            }
                        }
                    }
                    // if we did NOT get post-test data for user
                    else {
                        if (player.isInstructor) {
                            player.ds.generateTestScores('post');
                        }
                        else {

                            player.modal_dialog.show('Prerequisite Not Met',
                                'You must complete the ACES Post-Test before you can attempt this activity.',
                                [
                                    {
                                        text: "Close",
                                        click: function () {
                                            player.modal_dialog.close()
                                        }
                                    }
                                ],
                                {
                                    onEsc: function () {
                                        player.modal_dialog.close()
                                    }
                                }
                            );

                            /*
                            Standard_Dialog.open("You must complete the ACES Post-Test before you can attempt this activity.", {
                                title: 'Prerequisite Not Met'
                                //from: $('#ds_fail')
                            });
                            */
                            brbDisableActivity('Please exit out of this activity and complete the ACES Post-Test.', 'Post-Test not completed', 'No DS info found for user');
                        }
                    }
                },
                // failure
                function(a, b, c) {

                    if (typeof a === 'object') {
                        a = JSON.stringify(a);
                    } else if (a === undefined) {
                        a = '';
                    }
                    b = b === undefined ? '' : b;
                    c = c === undefined ? '' : c;

                    player.modal_dialog.show('System Error',
                        'There was an error retrieving your data. Please exit out of this activity and then open it again to resume.',
                        [
                            {
                                text: "Close",
                                click: function () {
                                    player.modal_dialog.close()
                                }
                            }
                        ],
                        {
                            onEsc: function () {
                                player.modal_dialog.close()
                            }
                        }
                    );

                    /*
                    Standard_Dialog.open("There was an error retrieving your data.", {
                        title: 'System Error'
                        //from: $('#ds_fail')
                    });
                    */
                    brbDisableActivity('Please exit out of this activity and then open it again to resume.', 'initializeUser failed for post-test', a, b, c);
                }
            );

        },

        generateTestScores: function(test) {

            //safe_log("generateTestScores: " + test);

            var userData;
            if (test === 'pre') {
                userData = player.ds.preUserData;
            }
            else {
                userData = player.ds.postUserData;
            }

            userData.scores = {};
            userData.scores.scores_new = {};

            /*
            for (var i=2; i<14; i++) {
                userData.scores.scores_new[i] = 10;
            }
            */

            if (test === 'pre') {
                userData.scores.scores_new [2] = 30;
                userData.scores.scores_new [3] = 20;
                userData.scores.scores_new [4] = 30;
                userData.scores.scores_new [5] = 20;
                userData.scores.scores_new [6] = 30;
                userData.scores.scores_new [7] = 20;
                userData.scores.scores_new [8] = 30;
                userData.scores.scores_new [9] = 20;
                userData.scores.scores_new [10] = 30;
                userData.scores.scores_new [11] = 20;
                userData.scores.scores_new [12] = 30;
                userData.scores.scores_new [13] = 20;
                userData.scores.scores_new [14] = 25;
                userData.scores.scores_new [15] = 25;
            }
            else {
                userData.scores.scores_new [2] = 30;
                userData.scores.scores_new [3] = 33;
                userData.scores.scores_new [4] = 20;
                userData.scores.scores_new [5] = 30;
                userData.scores.scores_new [6] = 29;
                userData.scores.scores_new [7] = 20;
                userData.scores.scores_new [8] = 30;
                userData.scores.scores_new [9] = 30;
                userData.scores.scores_new [10] = 28;
                userData.scores.scores_new [11] = 30;
                userData.scores.scores_new [12] = 32;
                userData.scores.scores_new [13] = 30;
                userData.scores.scores_new [14] = 35;
                userData.scores.scores_new [15] = 38;
            }

            //safe_log(userData);

            if (test === 'pre') {
                player.ds.preTestIsComplete = true;
                player.ds.instructorPreTestIsComplete = false;
            }
            else {
                player.ds.postTestIsComplete = true;
                player.ds.instructorPostTestIsComplete = false;
            }

        },

        // We are not using the callback function. Instead, we are listening
        // for both preTestIsComplete and postTestIsComplete to be set to
        // 'true' in initialize2, which will then call initialize3.
        initializeUser: function(callback) {

            // getPreTestData will call getPostTestData
            player.ds.getPreTestData();


        } // end initializeUser

    }, // end DS object

    initLiveFeedback: function() {
        //var live = (isIE || isMac) ? 'assertive' : 'polite';
        var live = "assertive";
        this.liveFeedback = $('<div id="live-feedback" class="visually-hidden" aria-live="' + live + '"></div>');
        $('body').append(this.liveFeedback);
    },

    setLiveText: function(text) {
        this.liveFeedback.text(text);
    },

    initialize: function() {

        this.modal_dialog = new ModalDialog();

        document.title = 'ACES Reflection';
        $('[data-type=chapter]').html('<main>' + $('[data-type=chapter]').html() + '</main>');
        $("h1.section-title").remove();

        // Add live feedback for readers
        this.initLiveFeedback();
        $("body").append("<div id='silenceElement' class='visually-hidden'></div>");

        // create <label>'s
        $('[data-tag="label"]').each(function() {
            $(this).changeElementType('label');
        });
        $('label').each(function() {
            $this = $(this);
            var data_for = $this.attr('data-for');
            if (data_for !== undefined) {
                $this.attr('for', data_for);
            }
        });

        // add classes
        $('[data-class]').each(function() {
            var $this = $(this);
            var data_class = $this.attr('data-class');
            $this.addClass(data_class);
        });

        // set scope for <th>'s
        $('th.scope_col, th.scope_row').each(function() {
            var $this = $(this);
            var scope = $this.attr('class');
            if (scope === "scope_col") {
                $this.attr('scope', 'col');
            }
            else if (scope === "scope_row") {
                $this.attr('scope', 'row');
            }
        });


        $('[data-block_type=fieldset]').each(function() {
            var $this = $(this);
            $this.attr('data-type', 'fieldset');
            $this.find("[data-type=title]").remove();

            // get rid of inner box
            var $inner_box = $this.children('div');
            var inner_box_html = $inner_box.html();
            $inner_box.remove();
            $this.html(inner_box_html);

            $this.changeElementType('fieldset');
        });

        $('[data-block_type=legend]').each(function() {
            var $this = $(this);
            var text = $this.find("p").html();
            var q_num = $this.attr('data-qnum');
            $this.html('<span class="visually-hidden">Question ' + q_num + ', </span>' + text);
            $this.changeElementType('legend');
        });

        $('[data-labelledby]').each(function () {
            var $this = $(this);
            var id = $this.attr('data-labelledby');
            $this.attr('aria-labelledby', id);
            $this.removeAttr('data-labelledby');
        });

        this._super();

        this.activity = new Activity_reflection();

    },

    show_toc: function () {
        $("#toc_modal_content").remove();
        var $menuDialog = $('<div>')
            .attr('id', 'toc_modal_content');
        $menuDialog
            .append(
                $('<p>').html(
                    '<strong>Note:</strong> A slide is not accessible until the preceding slide has been viewed or completed.')
            );

        var $menuList = $('<ul>');

        for (var i = 0; i < this.sections.length; ++i) {
            var s = this.sections[i];
            var extra_style = "";
            var isLinkEnabled = true;

            if (!player.isInstructor && i > this.last_available_section) {
                extra_style = "toc_section_not_available";
                //isLinkEnabled = false;
            }

            var $menuItem = $('<li>')
                .addClass(extra_style);

            /*
            var linkLabel = '';
            if (!isIE) {
                linkLabel += s.title;
                linkLabel += isLinkEnabled ? '' : ' unavailable';
            }
            */

            var ariaCurrent = (i === player.section_currently_showing) ? 'aria-current="page" ' : '';

            var $menuItemLink;
            if (player.isInstructor || i < player.last_available_section) {
                $menuItemLink = $('<a href="#slide' + i + '" tabindex="0" ' + ariaCurrent + 'onclick="player.toc_show_section(' + i + ')">')
                    .html('<span class="visually-hidden">Slide </span><span class="slide_num">' + (i + 1) + ': </span>' + s.title)
                    .addClass('toc_section_title');
            }
            else {
                $menuItemLink = $('<a tabindex="0" ' + ariaCurrent + '>')
                    .html('<span class="visually-hidden">Slide ' + (i + 1) + ', </span>' + s.title + '<span class="visually-hidden"> (not accessible)</span>')
                    .addClass('toc_section_title');
            }

            /*
            var $menuItemLink = $('<a ' + ariaCurrent + 'onclick="player.toc_show_section(' + i + ')">')
                .html('<span class="visually-hidden">Slide ' + (i + 1) + ', </span>' + s.title)
                .addClass('toc_section_title')
            */
            /*
            .attr({
                tabindex: 0,
                role: 'link',
                'aria-disabled': extra_style.indexOf('toc_section_not_available') !== -1
            });
            */
            //.attr('aria-label', linkLabel);

            $menuItem.append($menuItemLink);
            $menuList.append($menuItem);
        }

        $menuItemLink.addClass('last_slide');

        $menuDialog.append($menuList);

        $('#toc_modal').append($menuDialog).removeAttr('hidden');
        //$('#toc_modal_close_button').focus();
        $('body').addClass('toc-modal-open');

        player.setInert(true);

        if (viewingInPX()) {
            try {
                document.body.style.height = '100vh';
                parent.PX.API.Content.ResetHeight();
            }
            catch (e) {

            }
        }   

        document.getElementById('toc_modal_close_button').addEventListener('click', player.toc_modal_button_listener);
        document.body.addEventListener('click', player.toc_click_listener);
        document.body.addEventListener('keydown', player.toc_keydown_listener);

        setTimeout(function () {
            $('#toc_modal').focus();
        }, 500);

        /*
        Standard_Dialog.open($menuDialog[0].outerHTML, {
            from: $('#menu_button'),
            title: 'Table of contents'
        });

        $(".toc_section_title").each(function(i, v) {
            $(v).keypress(function(e) {
                if (( e.keyCode || e.which ) == 13) {
                    player.toc_show_section(i);
                    e.preventDefault();
                }
            });
        });
        */
    },

    setInert: function(set) {
        if (set) {
            $('main').attr('inert', 'true').attr('aria-hidden', 'true');
            $('header').attr('inert', 'true').attr('aria-hidden', 'true');
            $('nav').attr('inert', 'true').attr('aria-hidden', 'true');
        }
        else {
            $('main').removeAttr('inert').removeAttr('aria-hidden');
            $('header').removeAttr('inert').removeAttr('aria-hidden');
            $('nav').removeAttr('inert').removeAttr('aria-hidden');
        }
    },

    toc_keydown_listener: function(e) {
        var keycode = e.which;

        // Escape
        if (keycode === 27) {
            player.close_toc(true);
            console.log('setting focus on toc button');
            $('#menu_button').focus();
        }
        // Tab
        // If we tab out of the end of the list then go back to the close button
        else if (keycode === 9 && e.target.classList.contains('last_slide') && !e.shiftKey) {
            e.preventDefault();
            $('#toc_modal_close_button').focus();
        }
        // If we shift+tab from close button go to end of list
        else if (keycode === 9 && e.target.id === 'toc_modal_close_button' && e.shiftKey) {
            e.preventDefault();
            $('#toc_modal_content').find('.last_slide').focus();
        }
    },

    toc_click_listener: function(e) {
        console.log(e);
        if (!document.getElementById('toc_modal').contains(e.target) &&
            !document.getElementById('menu_button').contains(e.target)) {
            player.close_toc(true);
        }
    },

    toc_modal_button_listener: function(e) {
        e.stopPropagation();
        player.close_toc(true);
        return false;
    },

    close_toc: function(setFocusOnMenuButton) {
        //$('#menu_button').attr('aria-expanded', 'false');
        //$('#top_nav').removeClass('expanded');
        //$('#toc_dialog').remove();

        console.log('close_toc()');

        if (viewingInPX()) {
            try {
                document.body.style.height = 'unset';
                parent.PX.API.Content.ResetHeight();
            }
            catch (e) {

            }
        }

        player.setInert(false);
        $('body').removeClass('toc-modal-open');
        $('#toc_modal').attr('hidden', 'true');

        document.getElementById('toc_modal_close_button').removeEventListener('click', player.toc_modal_button_listener);
        document.body.removeEventListener('click', player.toc_click_listener);
        document.body.removeEventListener('keydown', player.toc_keydown_listener);

        if (setFocusOnMenuButton === true) {
            $('#menu_button').focus();
        }
    },

    toc_show_section: function(section_to_show) {
        Standard_Dialog.close();

        player.close_toc();

        player.show_section(section_to_show);
    },

    show_slide3: function() {

        //safe_log("show_slide3");

        var slide3_window = window.open("asset/slide3.html", "slide3", "toolbar=no,width=700");

    },

    show_section: function(section_to_show) {
        //this._super(section_to_show);

        // The following is the super show_section(). We are including it here and not
        // calling this._super(section_to_show) so we can change the behavior of the
        // pop-up shown to instructors

        console.log('show_section()');

        var direction = "right";
        if (section_to_show == "previous") {
            section_to_show = this.section_currently_showing - 1;
            direction = "left";
        }
        else if (section_to_show == "next") {
            section_to_show = this.section_currently_showing + 1;
        }

        if (section_to_show == null || isNaN(section_to_show) || section_to_show < 0 || section_to_show >= this.sections.length) {
            return;
        }

        // if user has to view sections in sequence and hasn't gotten up to this one, don't allow it
        if (this.md.sequenced_sections == "true" && section_to_show > this.last_available_section) {
            // but let instructors through
            if (player.activity.is_instructor()) {
                // brb: only show this pop-up once per session
                if (!player.INSTRUCTOR_NAV_MESG_SHOWN) {
                    player.POPUP_IS_SHOWING = true;

                    /*
                    Standard_Dialog.alert("Students need to read and complete each section of the activity before moving on to the next section. Instructors, however, can skip around between sections as they choose.",
                        {
                            from: $('#digfir_section_' + section_to_show).find('h2.section-title'),
                            diagTitle: 'Please note'
                        }
                    );
                    */

                    function close_dialog() {
                        //player.modal_dialog.close({ from: $('#digfir_section_' + section_to_show).find('h2.section-title') });
                        player.modal_dialog.close();
                        player.show_section(section_to_show);
                    }

                    var buttons = [
                        {
                            text: "OK",
                            click: close_dialog
                        }
                    ];
                    player.INSTRUCTOR_NAV_MESG_SHOWN = true;

                    console.log('show instructor note dialog');

                    player.modal_dialog.show('Please Note',
                        'Students need to read and complete each section of the activity before moving on to the next section. Instructors, however, can skip around between sections as they choose.',
                        buttons,
                        {
                            onEsc: close_dialog
                        }
                    );
                    return;
                }
            }
            else {
                player.POPUP_IS_SHOWING = true;
                Standard_Dialog.alert(this.md.section_sequence_message);
                return;
            }
        }
        else {
            player.POPUP_IS_SHOWING = false;
        }

        this.show_section_animate(section_to_show, direction);

        // process iframes for the section, unless all were preloaded
        if (player.md.preload_all_iframes != "true") {
            this.figures.process_iframes(this.get_current_section().jq);        // iframes
        }

        this.update_section_status();
        this.update_navigation();

        // end original super show_section()

        var curSection = this.section_currently_showing + 1;
        var secTitle = this.get_current_section().title;

        // 20210319: Don't announce slide number via live text
        //this.setLiveText('Slide ' + curSection + ': ' + secTitle + ' is now showing.');

        // 20210319: Set focus on h2 if this is not the initial viewing of the activity
        if (!player.INITIAL_VIEW) {
            $('#digfir_section_' + player.section_currently_showing).find('h2').focus();
        }
        else {
            player.INITIAL_VIEW = false;
        }

        // set focus on toc button
        // 20210319: We don't want to do this when the user initially loads the page
        if (!player.POPUP_IS_SHOWING) {
            //$("#menu_button").focus();
        }

        setTimeout(function () {
            try {
                console.log(parent.PX.API.Content.ResetHeight.toString());
                parent.PX.API.Content.ResetHeight();
            }
            catch (e) {

            }
        }, 500);

    },

    // This is only called once when activity originally loads
    show_navigation: function() {
        var header = "<header><h1>ACES Reflection: Compare and Contrast</h1></header><nav id='top_nav' aria-label='Table of contents'>"
            + "<button id='menu_button'>"
            + "<img src='asset/pics/toc_icon.gif' alt=''/><span id='menu_button_label' aria-hidden='true'>Outline</span><span class='visually-hidden'>Table of contents, opens dialog</span></button>"
            + "</nav>";

            $("[data-type=chapter]").prepend(header);

        // 20210319: This shouldn' be a nav (changed to div)
        var footer ="<nav aria-label='Slide'><div id='bottom_nav'>";

        // 20210319: These should be links, not buttons
        footer += player.create_prev_link();
        footer += "<div id='slide_num_display' aria-hidden='true'></div>"
        footer += player.create_next_link();
        footer += "</div></nav>";

        $("[data-type=chapter]").append(footer);

        // add click handler for toc button
        $('#menu_button').click(function (e) {
            var $this = $(this);

            player.show_toc();

            /*
            var expanded = $this.attr('aria-expanded');
            if (expanded === 'false') {
                $this.attr('aria-expanded', 'true');
                $this.parent('nav').addClass('expanded');
                player.show_toc();
            }
            else {
                $this.attr('aria-expanded', 'false');
                $this.parent('nav').removeClass('expanded');
                player.remove_toc();
            }
            */
        });

        // add click handlers for links
        $('#bottom_nav').click(function (e) {
            var id = e.target.id;
            if (id !== "next_button" && id !== "previous_button") {
                return;
            }
            e.preventDefault();
            if (id === "previous_button") {
                if (player.section_currently_showing !== 0) {
                    player.show_section('previous');
                }
            }
            else {
                if (player.section_currently_showing < player.sections.length - 1) {
                    player.show_section('next');
                } 
            }
        });

        // add div for toc modal
        $('body').prepend('<div id="toc_modal" tabindex="-1" aria-modal="true" role="dialog" aria-labelledby="toc_modal_heading" hidden><div class="heading"><h2 id="toc_modal_heading">Table of Contents</h2><button id="toc_modal_close_button"><span class="visually-hidden">Close</span><span aria-hidden="true">&#215;</span></button></div></div>');

    },

    create_prev_link: function () {
        var disabled = player.section_currently_showing === 0 ? true : false;
        
        var prev_link = '<a id="previous_button" href="#prev"><span class="visually-hidden">Previous slide</a>';
    
        return prev_link;
    },

    update_prev_link: function () {
        var $prevLink = $('#previous_button');
        $prevLink.attr('href', '#slide' + (player.section_currently_showing - 1));
        $prevLink.html('<span class="visually-hidden">Previous slide</a>')
    },

    create_next_link: function () {
        //var disabled = player.section_currently_showing === player.last_available_section ? true : false;

        var next_link = '<a id="next_button" href="#next"><span class="visually-hidden">Next slide</a>';
        return next_link;
    },

    update_next_link: function() {
        var $nextLink = $('#next_button');
        $nextLink.attr('href', '#slide' + (player.section_currently_showing + 1));
        $nextLink.html('<span class="visually-hidden">Next slide</a>');
    },

    show_section_animate: function(section_to_show, direction) {
        this._super(section_to_show, direction);

        // update aria-hidden attribute on sections
        var curr_id = 'digfir_section_' + section_to_show;

        $('[data-type="section"]').each(function() {
            var $this = $(this);
            var id = $this.attr('id');

            if (id === curr_id) {
                $this.attr('aria-hidden', 'false');
            } else {
                $this.attr('aria-hidden', 'true');
            }

        });

        // update Previous/Next buttons
        if (this.section_currently_showing === 0) {
            $("#previous_button").addClass("disabled")
                .attr('aria-hidden', 'true')
                .attr('tabindex', '-1');
        }
        else {
            $("#previous_button").removeClass("disabled")
            .removeAttr('aria-hidden')
            .removeAttr('tabindex');
        }

        if (this.section_currently_showing == (this.sections.length - 1)) {
            $("#next_button").addClass("disabled")
                .attr('aria-hidden', 'true')
                .attr('tabindex', '-1');
        }
        else {
            $("#next_button").removeClass("disabled")
                .removeAttr('aria-hidden')
                .removeAttr('tabindex');
        }

        $('#slide_num_display').text("Slide " + (this.section_currently_showing + 1) + " of " + this.sections.length);
        //var snd_width = $('#slide_num_display').width();
        //$('#slide_num_display').css({'left': 'calc(50% - ' + (snd_width/2) + 'px)'});

        player.update_prev_link();
        player.update_next_link();
    },

    initialize2_super: undefined,

    // 20210319: Added this method to allow us to add slide number to h2
    initialize_sections: function () {
        var player = this;	// define this so it's available inside the below function
        $("[data-type='section']").each(function (index, element) {
            var jq = $(element);

            // establish the section in the sections array; store its jq
            var s = player.sections[player.sections.length] = new Object();
            s.jq = jq;

            // set the section's index
            s.index = index;
            jq.attr("data-section-index", index).attr("id", "digfir_section_" + index);

            // extract the section's title and number
            s.number = jq.find("h2").find("[data-type=number]").html();
            s.title = jq.find("h2").find("[data-type=title]").html();

            var $h2 = jq.find("h2");
            $h2.html('<span class="visually-hidden">Slide ' + (index + 1) + ', </span>' + s.title);
            $h2.attr('tabindex', '-1');

            // process page queries for the section
            player.activity.process_page_queries_for_section(jq);
            ++player.section_count;
        });
    },

    initialize2: function () {

        //safe_log("initialize2...");

        this.initialize2_super = this._super;

        if (player.ARGA_running) {
            try {
                var dejs_arr = ARGA_API.dejs_data;
                for (var i=0; i<dejs_arr.length; i++) {
                    if (dejs_arr[i].name === "cmi.completion_status") {
                        this.ACTIVITY_COMPLETED = (ARGA_API.dejs_data[i].value == "completed");
                        //safe_log("ACTIVITY_COMPLETED = " + this.ACTIVITY_COMPLETED);
                        break;
                    }
                }
            }
            catch(err) {
                safe_log("initialize2: Can't get cmi.completion_status: " + err)
            }
        }

        console.log('ARGA_running = ' + player.ARGA_running);

        if (player.ARGA_running) {
            if (Get_ARGA_Data('user_rights') == '3_instructor') {
                player.isInstructor = true;
            }
        }
        // If we aren't in LP then make user an instructor so we
        // can view the activity
        else {
            player.isInstructor = true;
        }

        safe_log("isInstructor = " + player.isInstructor);

        // create h3 tags
        $('[data-block_type="h3"]').each(function() {
            var $this = $(this);
            var text = $this.find('p').html();
            $this.html(text);
            $this.changeElementType('h3');
        });

        // get rid of all question h3 headers
        $('[data-type="question"] > h3').remove();

        // change <p> to <div> in <fieldset>
        $('fieldset').find('p').each(function () {
            $(this).changeElementType('div');
        });

        // get rid of empty [data-type="query_display"] in <fieldset>
        $('fieldset').find('[data-type="query_display"]').each(function () {
            $(this).remove();
        });

        /*
        $('.question_action_div').each(function () {
            var $this = $(this);
            console.log('found question_action_div: ', $this);
            var id = $this.attr('id');
            var fb_id = id.replace(/action_div/, 'feedback');
            $this.prepend('<div class="question_feedback" role="alert" tabindex="-1" id="' + fb_id + '"></div>');
        });
        */

        // add question number to legend
        /*
        $('[data-block_type=legend]').each(function () {
            var $this = $(this);
            var $parent = $this.parent().parent();
            console.log($parent.attr('data-question_index'));
            var q_num = $parent.attr('data-question_index');
            var q_html = $this.html();
            $this.html('<span class="visually-hidden">Question ' + q_num + ', </span>' + q_html);
        });
        */

        // Get DS saved info for user. If the user has not completed either
        // the pre/post-test then this will terminate the activity and
        // present an error message to the user.
        player.ds.initialize();

        // wait until both pre/post-test data has been loaded and then
        // continue with initialization
        //safe_log("waiting for DS student data to load...");
        var timer = setInterval(function() {
            if (player.ds.preTestIsComplete && player.ds.postTestIsComplete) {
                clearInterval(timer);
                //safe_log("DS student data loaded, calling initialize3");
                player.initialize3();
            }
            else if (player.ds.abortActivity) {
                clearInterval(timer);
                safe_log("activity aborted");
            }
        }, 250);


    }, // end initialize2


    // This will be called by getPostTestData if the user has
    // completed both pre and post test
    initialize3: function() {

        //safe_log("initialize3...");

        //safe_log("platform name: " + platform.name);
        //safe_log("platform OS family: " + platform.os.family);


        //safe_log(player.ds.preUserData);
        //safe_log(player.ds.postUserData);

        this.initialize2_super();

        $("h2.section-title").show();

        $('label.select').each(function() {
            $(this).parent('span').addClass('select_span');
        });

        $("[data-block_type='comparison_chart']").each(function() {
            $(this).html(player.createComparisonChartHtml());

            if (!player.ds.instructorPreTestIsComplete ||
                !player.ds.instructorPostTestIsComplete) {
                var message = "<p><strong>Instructor Note:</strong> Since you have not completed ";
                if (!player.ds.instructorPreTestIsComplete && !player.ds.instructorPostTestIsComplete) {
                    message += "ACES, scores in both columns have been randomly generated for you to illustrate the student experience. Students will see their own scores in this table when they complete the activity.";
                }
                else if (!player.ds.instructorPreTestIsComplete) {
                    message += "the ACES Pre-Test, the scores for the Initial Report column have been randomly generated for you to illustrate the student experience. Students will see their own scores in this table when they complete the activity.";
                }
                else {
                    message += "the ACES Post-Test, the scores for the Progress Report column have been randomly generated for you to illustrate the student experience. Students will see their own scores in this table when they complete the activity.";
                }
                $(this).before(message);
            }
        });

        $("#print_summary").append("<a "
            + "href='#' onclick='player.printSummary();'>"
            + "Print Summary <span class='visually-hidden'>, "
            + "Opens in new window</span></a>");
    },

    fix_local_links: function() {
        $('[data-block_type=local_link]').each(function() {
            var $this = $(this);
            $this.changeElementType('a');
        });

        $('[data-block_type=local_link]').each(function() {
            var $this = $(this);
            var section = $this.attr('data-section');
            //$this.attr('tabindex', '0');
            $this.attr('href', '#section' + section);
            //safe_log($this.get(0));

            var link_text = $this.text();
            $this.html(link_text + "<span class='visually-hidden'>, Opens in new window</span>");
            //$this.attr('title', 'Link opens in new window');

            $this.get(0).onclick = function() {
                //player.toc_show_section(parseInt(this.getAttribute("data-section")));
                player.show_slide3();
                return false;
            }
        });

        /*
        $('[data-block_type=fieldset]').each(function() {
            var $this = $(this);
            $this.attr('data-type', 'fieldset');
            $this.find("[data-type=title]").remove();
            $this.changeElementType('fieldset');
        });
        */

        /*
        $('[data-block_type=legend]').each(function() {
            var $this = $(this);
            var text = $this.find("p").html();
            $this.html(text);
            $this.changeElementType('legend');
        });
        */
    },

    createComparisonChartHtml: function() {
        var html = "<table class='comparison_chart'><caption>ACES Comparison Chart</caption><tbody>"
            + "<tr><th scope='col'>Scale</th><th scope='col'>Score on Initial Report</th><th scope='col'>Score on Progress Report</th></tr>";

        for (var i=0; i<player.measure_names_new.length; i++) {

            var measure_key = Number(player.measure_names_new[i].key);
            var measure_name = player.measure_names_new[i].value;

            html += "<th scope='row'>" + measure_name + "</th><td>" + player.scale_percentages_new[measure_key][player.ds.preUserData.scores.scores_new[measure_key]] + "%</td><td>" + player.scale_percentages_new[measure_key][player.ds.postUserData.scores.scores_new[measure_key]] + "%</td></tr>";
        }

        html += "</tbody></table>";

        return html;
    },

    // This method returns the html for the question answer so it
    // can be used in pop-up window.
    getQuestionAnswer: function(question_number) {
        var question = player.activity.questions[question_number];
        if (!question.is_complete()) {
            return "<p>Question not answered.</p>";
        }

        // if we have more than one query then this is a list of CB's
        if (question.queries.length > 1) {
            var html = "<div data-block_type='fieldset'>";
            var answers = "";

            // go through each query (CB) and if checked add it to answers
            for (var i = 0; i<question.queries.length; i++) {
                var answer = question.queries[i].getHTML('review');
                if (!/Not Checked/.test(answer)) {
                    answer = answer.replace(/<span type="checkbox" .*/, "");
                    answers += "<li>" + answer + "</li>";
                }
            }

            if (answers === "") {
                answers = "<p>No ACES scores were checked for this question.</p>";
            }
            else {
                answers = "<ul>" + answers + "</ul>";
            }

            html += answers + "</div>";
            return html;
        }
        // otherwise this is an MC or Essay question and we just have one query
        else {
            var answer = question.queries[0].getHTML('review');
            answer = answer.replace(/visually-hidden/, "");
            console.log(answer);
            var html = answer;
            if (question.queries[0].getQueryTypeForARGA() === "essay") {
                html = "<h3>Question</h3>" + html;
            }
            return html;
        }
    },

    printSummary: function() {
        var win = window.open('asset/summary.html', 'printwindow');
    }

});

player = new Player_reflection();

// To be called if the connection to DS doesn't work.
// error_msg should always be a non-empty string
// a, b, c are optional additional error messages
function brbDisableActivity(error_mesg, a, b, c, d) {

    player.ds.abortActivity = true;

    var additional_errors = "";

    if (a !== undefined) {
        if (typeof a === 'object') {
            a = JSON.stringify(a);
        }
        additional_errors += '<br/>' + a;
    }
    if (b !== undefined) {
        if (typeof b === 'object') {
            b = JSON.stringify(b);
        }
        additional_errors += '<br/>' + b;
    }
    if (c !== undefined) {
        if (typeof c === 'object') {
            c = JSON.stringify(c);
        }
        additional_errors += '<br/>' + c;
    }
    if (d !== undefined) {
        if (typeof d === 'object') {
            d = JSON.stringify(d);
        }
        additional_errors += '<br/>' + d;
    }

    // brb: they do not want to show technical errors to users

    if ($('#ds_fail').length < 1) {
        //$("#manuscript").html("<p id='ds_fail'>" + error_mesg + "</p><p>" + additional_errors + "</p>");
        $("#manuscript").html("<p id='ds_fail'>" + error_mesg + "</p>");
    }
    /*
    else {
        $("#manuscript").append("<p>" + error_mesg + "</p>");
    }
    */

    //window.parent.PX.API.DataService.setData = window.parent.PX.API.DataService.setDataOrig;

    //$('#ds_fail').attr('tabindex', '-1');

    player.setLiveText(error_mesg);
}


Essay_Query = Essay_Query.extend({
    focus_user_input: function() {
        // Empty by design (overload default method)
    },

    // brb: override so we can clean the query text sent to the grade book
    getQueryTextForARGA: function() {
        if (this.query_text_for_ARGA != null && this.query_text_for_ARGA != "") {
            return this.query_text_for_ARGA;
        } else if (this.query_text != null && this.query_text != "") {
            return player.clean_query_text(this.query_text);
        } else {
            return "see page for question context";
        }
    },

    // this is called by question.is_complete() to determine if question has been completed
    // since we can have a min word limit for essay questions, we can't just check if the user
    // has entered text, we must check that they entered enough text to get a grade of 100
    userHasAnswered: function() {
        //return (this.user_answer != "" && this.user_answer != null);
        return (this.grade == 100);
    },

    sanitize_user_input: function(text) {
        text = text.replace(/&/g, "&amp;");
        text = text.replace(/</g, "&lt;");
        text = text.replace(/>/g, "&gt;");
        text = text.replace(/"/g, "&quot;");
        text = text.replace(/'/g, "&#x27;");
        text = text.replace(/\//g, "&#x2F;");

        return text;
    },

    evaluateAnswer: function(question_index) {
	// if user_answer_is_submittable is true, we're already ready to proceed
	if (this.user_answer_is_submittable == true) {
	    return true;
	}

	// otherwise, get the user's answer
	var raw_answer = $.trim($("#query_answer_" + this.query_index).val());

        console.log('raw answer: ' + raw_answer);

        //safe_log("Essay_Query.evaluateAnswer");
        //safe_log(raw_answer);
        //safe_log("sanitizing...");

        var sanitized_answer = this.sanitize_user_input(raw_answer);

        //safe_log(sanitized_answer);

        //safe_log("fixing new lines");
	// replace newlines with breaks in recorded answer
	//this.user_answer = sanitized_answer.replace(/\n/g, "<br/>");
        this.user_answer = sanitized_answer;
        //safe_log(this.user_answer);

	// if answer is blank and author didn't explicitly say to accept answers, make them enter something.
	if (raw_answer == "" && this.md.accept_blank_answers != "true") {
	    this.set_evaluation_feedback("You must enter an answer.");
	    //this.focus_user_input();
	    // return false, to say that we can't yet evaluate the answer.
            player.setLiveText("You must enter an answer.");
	    return false;

	    // else check with MAVE if specified
	} else if (this.md.mave_question_id != null) {
	    var t = this;

	    MAVE.validate_answer(this.md.mave_question_id, raw_answer, {
		callback_fn:function(x) {
		    t.MAVE_callback(x);
		}
		, max_similarity_score: this.md.max_similarity_score
		, min_valid_score: this.md.min_valid_score
		, max_question_text_score: this.md.max_question_text_score
		, validity_criterion: this.md.validity_criterion
		, plagiarism_criterion: this.md.plagiarism_criterion
	    });

	    // store the question_index, so we'll be able to re-call player.submit_question on callback
	    this.question_index = question_index;

	    // return false, to say that we can't yet evaluate the answer.
	    return false;

	    // otherwise evaluate here...
	} else {
	    // if we have a min word count, compare to it
        // 20210405: Not doing min word count check any more since we are
        // now confirming that user wants to submit essay answer.
        /*
	    if (this.md.min_word_count != null) {
            console.log('checking unique word count for: ' + this.user_answer);
		    var word_count = this.unique_word_count(this.user_answer);
		    if (word_count < this.md.min_word_count * 1) {
		        this.set_evaluation_feedback("Your answer is not long enough. It must be at least 10 unique words. Please elaborate and resubmit.");
                player.setLiveText("Your answer is not long enough. It must be at least 10 unique words. Please elaborate and resubmit.");
		        return false;
		    }
	    }
        */
	    // If we make it to here, assume anything is OK.
	    this.setGrade(100);
	    this.answer_was_just_entered = true;
	    this.user_answer_is_submittable = true;
	    return true;
	}
    },
    user_has_answered: function() {
        if ($.trim($("#query_answer_" + this.query_index).val() != "")) {
            return true;
        } else {
            return false;
        }

    },
    query_text_html: function(mode) {

        if (this.query_text != "" && this.md._hide_query_text != "true") {
            // replace newlines with <br />
            var qt = jQuery.trim(this.query_text).replace(/\n/g, "<br>");

            var underlines = qt.match(/_{2,20}/g);

            if (underlines) {
                $(underlines).each(function(i, value) {
                    var newAriaString = '<span aria-hidden="true">' + value + '</span><span class="visually-hidden">blank</span>';
                    qt = qt.replace(value, newAriaString);
                });
            }

            /*
            if (mode === "review") {
                return "<p class='query_text' id='textarea_description_question_"
                    + this.query_index + "'>" + qt
                    + "</p>";
            }
            */

            /*
            return "<label for='query_answer_" + this.query_index
                + "' class='visually-hidden'>Essay question</label><legend class='query_text' id='textarea_description_question_"
                + this.query_index + "'>" + qt
                + "</legend>";
            */
            return "<label for='query_answer_" + this.query_index + "'><span class='visually-hidden'>Question " + this.parent_question.jq.attr('data-question_index') + ", </span>" + qt + "</label>";
        }
        else {
            return "";
        }
    },
    getHTML: function(mode) {
        // start with the query text, if there
        var html = this.query_text_html(mode);

        var val, disabled;
        if (!this.userHasAnswered() || mode == 'preview') {
            val = "";
            disabled = "";

            // otherwise we have an answer, so fill it in
        }
        else {
            // make sure we convert double quotes to &#34; for placement in 'value="xxx"' attributes
            val = this.user_answer;

            // convert breaks to newlines
            //val = val.replace(/<br *(\/)?>/g, "\n");

            // also disable the input if we're not in review_correct_incorrect mode
            if (mode != "review_correct_incorrect") {
                disabled = "disabled";
            }
        }

        /*
        if (mode === "review") {
            html += "<h3 class='essay_question_answer'>Your Answer</h3>";
            html += "<div id='query_answer_" + this.query_index + "' class='query_essay_ta query_essay_answer' aria-describedby='textarea_description_question_" + this.query_index + "'>";
            html += val;
            html += "</div>";
        }
        */
        //else {
            /*
            html +=
                "<textarea id='query_answer_" + this.query_index + "' class='query_essay_ta' " + disabled
                + " aria-describedby='textarea_description_question_" + this.query_index + "'>"
                + val + "</textarea>";
            */
            html +=
                "<textarea id='query_answer_" + this.query_index + "' class='query_essay_ta' " + disabled
                + ">"
                + val + "</textarea>";

            // if answer was just entered and grade is 100, also show the "provisional" message
            // (unless for this question it isn't provisional)
            if (this.answer_was_just_entered && this.grade == 100 && this.md.points_are_provisional == "true") {
                html +=
                    "<div class='essay_question_evaulation_feedback' id='essay_question_evaulation_feedback_" + this.query_index
                    + "'>Your answer has been provisionally accepted.  You will receive full credit at this time, but your instructor may update your grade later after evaluating your answer.</div>";
            }

        //}

        return html;
    }

});


DD_Query = DD_Query.extend({

    parseSource: function() {
	// initialize vars
	this.choices = new Array();
	this.correct_answer_index = null;
	this.user_answer_index = null;

	// DD's are always one line only -- if query_text is specified, it's via metadata (see below)
	this.query_text = "";

	// the line might start with "DD:"; if so, strip it out
	var s = jQuery.trim(this.query_source.replace(/^DD:\s*/, ""));

        //safe_log("s = [" + s + "]");

	// if line ends with "~~\d+", that's points
	if (s.search(/(.*)~~([\.\d]+)$/) > -1) {
	    s = RegExp.$1;
	    this.points_possible = RegExp.$2;
	}

        // if line ends with ##text..., that's the label
        if (s.search(/(.*)##(.*)$/) > -1) {
	    s = RegExp.$1;
	    this.label = RegExp.$2;
	}

        //safe_log("s = [" + s + "]");

	// if the line starts with "!!", don't scramble
	if (s.substr(0, 2) == "!!") {
	    this.md.never_scramble = "true";

	    // strip the !! out of answer
	    s = s.replace(/^\!\!/, "");
	}

        //safe_log("s = [" + s + "]");

	// if the line ends with "??", any answer is correct
	if (s.substr(-2, 2) == "??") {
	    this.md.no_correct_answer = "true";

	    // strip the ?? out of answer
	    s = s.replace(/\?\?$/, "");
	}

        //safe_log("s = [" + s + "]");

	// if any commas were sent in as "\,", turn them into &#44;
	s = s.replace(/\\,/g, "&#44;");

	// likewise, turn "\*" into &#42;
	s = s.replace(/\\\*/g, "&#42;");

	// we might have &#34;'s indicating double quotes in the answers;
	// convert these to actual double quotes so that they display properly (though it might not matter)
	s = s.replace(/\&#34;/g, '"');

	// answer string should be something like "choice 1, choice2, *correct choice"
	// so split it and find the correct answer
	var answers = s.split(/,\s*/);
	var correctChoice = -1;
	for (var i = 0; i < answers.length; ++i) {
	    var ans = answers[i];

	    // look for points...
	    if (ans.search(/^_points:\s*(\d+)/i) > -1) {
		this.points_possible = RegExp.$1;

		// skip to the next answer
		continue;

		// or query text
	    } else if (ans.search(/^_query_text:\s*(.*)/i) > -1) {
		this.query_text = RegExp.$1;

		// skip to the next answer
		continue;


		// or metadata -- starts with "_[a-zA-Z]"
	    } else if (ans.search(/^\_[a-zA-Z]/) > -1) {
		this.extractMetadata(ans);

		// skip to the next answer
		continue;

		// if we've found the correct answer...
	    } else if (ans.substr(0,1) == "*") {
		// if we already got a correctChoice, there's an error
		if (this.correct_answer_index != null) {
		    return "<b>error in pulldown answer (too many correct answers): " + s + "</b>";
		}

		// otherwise this one is correct
		this.correct_answer_index = i;

		// strip the leading *
		ans = ans.substr(1);
	    }

	    // convert &#42;s to literal *'s now; also convert commas
	    ans = ans.replace(/\&#42;/g, '*');
	    ans = ans.replace(/\&#44;/g, ',');

	    // transfer to choices, trimming in the process
	    this.choices.push(jQuery.trim(ans));
	}

	// note: answers will be coded 0, 1, 2, 3, etc., with correctChoice numbered accordingly
	// (e.g. if first answer is correct, correctChoice is 0)

	// if we didn't get a correct choice, or if there aren't enough items,
	// that's an error
	if (this.correct_answer_index == null) {
	    return "<b>error in pull down answer (no correct answer specified): " + s + "</b>";
	} else if (this.choices.length <= 1) {
	    return "<b>error in pulldown answer (0 or 1 choices specified): " + s + "</b>";
	}

	// shuffle choices, unless md.never_scramble is true
	this.choice_order = new Array();
	for (var j = 0; j < this.choices.length; ++j) this.choice_order[j] = j;
	if (this.md.never_scramble != "true") {
	    this.choice_order.shuffle();
	}

	// deal with error handling...
    },

    getHTML: function(mode) {

        //safe_log("DD_query.getHTML: mode = " + mode);
        //safe_log(this);

	var html = "";

	// DD's never have query text
	// Start a nobr tag for the item
	html = '<span class="query_dd_select_span ' + mode + '">';

        if (this.label !== null && this.label !== undefined) {
            if (mode !== "review") {
                html = '<label class="select" for="query_answer_' + this.query_index + '">' + this.label + '</label>' + html;
            }
            else {
                html = '<span class="select_label">' + this.label + '</span>' + html;
            }
        }


	// if we're in preview mode or
	// (a student answer has been given and we're not in review_correct_incorrect mode),
	// or the student got it correct
	// disable the select
	var disabled = "";
	if (mode == 'preview' || (mode != 'review_correct_incorrect' && this.user_answer_index != null) || this.isCorrect()) {
	    disabled = "disabled";
	}

        if (mode !== "review") {
	    html += '<select class="query_dd_select" id="query_answer_' + this.query_index + '" ' + disabled + '>'
	        + '<option value="-1">Not selected</option>'
        }

	for (var j = 0; j < this.choices.length; ++j) {
	    var index = this.choice_order[j];
	    var selected = "";
	    // select the student answer, unless we're in preview mode
	    if (this.user_answer_index == index && mode != 'preview') {
		selected = "selected"
                if (mode === "review") {
                    html += this.choices[index];
                }
	    }

            if (mode !== "review") {
	        html += '<option value="' + index + '" ' + selected + '>' + this.choices[index] + '</option>';
            }
	}
        if (mode !== "review") {
	    html += '</select>'
        }

	html += this.finish_inline_query(this.choices[this.correct_answer_index], mode);

	// Close off the nobr tag for the item
	html += '</span>';

	return html;
    },

    evaluateAnswer: function() {
	// get the user's selected answer, which will be 0-based
	this.user_answer_index = $('#query_answer_' + this.query_index).val() * 1;

        //safe_log("DD_Query.evaluateAnswer: user_answer_index = " + this.user_answer_index);

        // if we didn't choose an answer, then return false
        if (this.user_answer_index < 0) {
            this.set_evaluation_feedback("You must make a selection for each item.");
	    //this.focus_user_input();
	    // return false, to say that we can't yet evaluate the answer.
            player.setLiveText("You must make a selection for each item.");

            return false;
        }

        var parent_index = this.parent_question.index;
        $('#dd_question_evaulation_feedback_' + parent_index).remove();

	// set grade accordingly (if this.md.no_correct_answer is true, any answer is correct)
	if (this.user_answer_index == this.correct_answer_index || this.md.no_correct_answer == "true") {
	    this.setGrade(100);
	} else {
	    this.setGrade(0);
	}

	return true;
    },

    set_evaluation_feedback: function(html) {

        //safe_log("DD_Query.set_evaluation_feedback");
        //safe_log(this);

        var parent_index = this.parent_question.index;
        var $parent = $('[data-question_index="' + parent_index + '"]');
        var $dd_eval_div = $('#dd_question_evaulation_feedback_' + parent_index);

        if ($dd_eval_div.length > 0) {
            $dd_eval_div.html(html);
        }
        else {
            html = "<div class='dd_question_evaulation_feedback' id='dd_question_evaulation_feedback_" + parent_index + "'>" + html + "</div>";
            $parent.find('#question_action_div_' + parent_index).before(html);
        }
    }

});

MC_Query = MC_Query.extend({

    query_text_html: function() {
        if (this.query_text != "" && this.md._hide_query_text != "true") {
            // replace newlines with <br />
            var qt = jQuery.trim(this.query_text).replace(/\n/g, "<br>");

            var underlines = qt.match(/_{2,20}/g);

            if (underlines) {
                $(underlines).each(function(i, value) {
                    var newAriaString = '<span aria-hidden="true">' + value + '</span><span class="visually-hidden">blank</span>';
                    qt = qt.replace(value, newAriaString);
                });
            }

            return "<legend class='query_text' id='description_question_" + this.query_index + "'><span class='visually-hidden'>Question " + this.parent_question.jq.attr('data-question_index') + ", </span>" + qt + "</legend>"
        }
        else {
            return "";
        }
    },

    getHTML: function(mode) {

        //safe_log("MC_Query.getHTML: mode = " + mode);
        //safe_log(this);

        // initialize imagemap elements
        if (this.imagemap_initialized === false) {
            var question_jq = this.jq.parent();
            this.im_init(question_jq);
            this.imagemap_initialized = true;
        }

        var fs_disabled = (mode === 'review') ? ' disabled' : '';

        var decrId = "description_question_" + this.query_index;
        var html = "";
        /*
        if (mode === "review") {
            html += "<h3 class='visually-hidden'>Statement</h3><p class='query_text' id='description_question_" + this.query_index + "'>" + this.query_text + "</p><div class='mc_query_group'>";
        }
        */
        //else {
            html += /*"<h3 class='visually-hidden'>Statement</h3>*/"<fieldset" + fs_disabled + ">" + this.query_text_html() + "<div class='mc_query_group'>";
        //}

        // Build answer selection
        var is_correct = false;
        for (var z = 0; z < this.choices.length; z++) {
            var index = this.choice_order[z];
            // if this is review mode...
            var checked = "";
            var disabled = "";
            var radio_style = "query_mc_other_choice";
            var label_class = "mc_label";
            //var label_tag = (mode === "review") ? "span" : "label";
            var label_tag = "label";
            
            if (mode == 'review' || mode == 'review_correct_incorrect' || mode == 'preview') {

                if (this.md.no_correct_answer === "true") {
                    if (index == this.user_answer_index) {
                        is_correct = true;
                        label_class += " user_answer";
                    }
                }
                else {
                    // if it's the correct choice, it should be marked in green
                    if (index == this.correct_answer_index) {
                        // but only if we're *not* in review_correct_incorrect mode or this is the user's choice
                        // if (mode != 'review_correct_incorrect' || index == this.user_answer_index) {
                        radio_style = "query_mc_correct_choice";
                        // }

                        // if this is what the user chose, he got it correct
                        if (index == this.user_answer_index) {
                            is_correct = true;
                        }

                        // else if it's the user's choice (and this.md.no_correct_answer is not true), it should be marked in red
                    } else if (index == this.user_answer_index) {
                        radio_style = "query_mc_incorrect_user_choice";
                    }

                }
                // and the input should be disabled, unless we're in review_correct_incorrect mode
                if (mode != "review_correct_incorrect") {
                    disabled = "disabled";
                }
            }

            // if this is the choice the user clicked, it's checked (regardless of mode)
            if (index == this.user_answer_index) {
                checked = "checked";
            }

            /*
            if (mode === "review") {
                // we are only going to show the selected answer
                if (checked === "checked") {
                    html += '<div id="query_answer_' + this.query_index  + '" class="mc_option answered">';
                    html += '<h3>Your Rating</h3>';
                    html += '<p class="user_answer">' + this.choices[index]  + '</p>';
                    html += '</div>';

                }
            }
            */
            //else {
            
                html += "<span class='mc_option'><span><span class='" + radio_style + "'>"
                    //+ "<input class='query_mc_choice_radio' aria-describedby='" + decrId
                    + "<input class='query_mc_choice_radio"
                    + "' type='radio' name='query_answer_" + this.query_index
                    + "' "
                    + " id='query_answer_" + this.query_index + "_" + index + "' "
                    + " value='" + index + "' "
                    + checked + " "
                    + disabled + " "
                    + " /> "
                    + "</span>"
                    + "<" + label_tag + " class='" + label_class + "' id='query_answer_label_" + this.query_index + "_" + index + "' "
                    + " for='query_answer_" + this.query_index + "_" + index + "'>"
                    + this.choices[index]
                    + "</" + label_tag + ">"

                    + "</span></span>"
                ;
            //}
        }

        // get a possible feedback string; this fn will take account of mode
        html += this.correct_incorrect_feedback(mode, is_correct);

        if (mode === "review") {
            html += "</div>";
        }
        else {
            html += "</div></fieldset>";
        }

        return html;
    },

    evaluateAnswer: function() {

        if (!this.user_has_answered()) {
            this.set_evaluation_feedback("You must select an answer.");
            player.setLiveText("You must select an answer.");
            return false;
        }

        var parent_index = this.parent_question.index;
        $('#mc_question_evaulation_feedback_' + parent_index).remove();

	// get the user's selected answer, which will be 0-based
	this.user_answer_index = $('input:radio[name=query_answer_' + this.query_index + ']:checked').val() * 1;

	// set grade accordingly (if this.md.no_correct_answer is true, any answer is correct)
	if (this.user_answer_index == this.correct_answer_index || this.md.no_correct_answer == "true") {
	    this.setGrade(100);
	} else {
	    this.setGrade(0);
	}

	return true;
    },

    set_evaluation_feedback: function(html) {

        //safe_log("MC_Query.set_evaluation_feedback");
        //safe_log(this);

        var parent_index = this.parent_question.index;
        $('#question_error_' + parent_index).text(html);

        /*
        var $parent = $('[data-question_index="' + parent_index + '"]');
        var $mc_eval_div = $('#mc_question_evaulation_feedback_' + parent_index);

        if ($mc_eval_div.length > 0) {
            $mc_eval_div.html(html);
        }
        else {
            html = "<div class='mc_question_evaulation_feedback' id='mc_question_evaulation_feedback_" + parent_index + "'>" + html + "</div>";
            $parent.find('#question_action_div_' + parent_index).before(html);
        }
        */
    },

    user_has_answered: function() {

        if (typeof $('input:radio[name=query_answer_' + this.query_index + ']:checked').val() != "undefined") {
            return true;
        }
        else {
            return false;
        }
    }

});


CB_Query = CB_Query.extend({

    getQueryTextForARGA: function() {

        safe_log("MC_Query.getQueryTextForARGA()");
        safe_log(this);

        var query_number = '' + this.query_index;

        if (query_number in CB_QUERY_TEXT) {
            return CB_QUERY_TEXT[query_number];
        }
        else {
            return "see page for question context";
        }
    },

    removeLabel: function() {
        //safe_log("removeLabel");
        //safe_log(this);

        var query_index = "query_answer_" + this.query_index;

        var $label = $("label[data-for=" + query_index + "]");
        if ($label.length > 0) {
            $label.changeElementType('span');
        }

        if (this.user_answer_index == 0) {
            $("span[data-for=" + query_index + "]").addClass("checked");
        }
        else {
            $("span[data-for=" + query_index + "]").addClass("unchecked");
        }
    },

    parseSource: function() {

        //safe_log("CB.parseSource()");

        // initialize vars
        this.choices = new Array();
        this.correct_answer_index = null;
        this.user_answer_index = null;

        // CB's are always one line only -- if query_text is specified, it's via metadata (see below)
        this.query_text = "";

        // the line might start with "CB:"; if so, strip it out
        var s = jQuery.trim(this.query_source.replace(/^CB:\s*/, ""));

        // if line ends with "~~\d+", that's points
        if (s.search(/(.*)~~([\.\d]+)$/) > -1) {
            s = RegExp.$1;
            this.points_possible = RegExp.$2;
        }

        // if line ends with "##text...", that's the label
        if (s.search(/(.*)##(.*)$/) > -1) {
            s = RegExp.$1;
            this.label = RegExp.$2;
        }

        // if the line ends with "??", any answer is correct
        if (s.substr(-2, 2) == "??") {
            this.md.no_correct_answer = "true";

            // strip the ?? out of answer
            s = s.replace(/\?\?$/, "");
        }

        // if any commas were sent in as "\,", turn them into &#44;
        s = s.replace(/\\,/g, "&#44;");

        // likewise, turn "\*" into &#42;
        s = s.replace(/\\\*/g, "&#42;");

        // we might have &#34;'s indicating double quotes in the answers;
        // convert these to actual double quotes so that they display properly (though it might not matter)
        s = s.replace(/\&#34;/g, '"');

        // answer string should be something like "choice 1, *correct choice"
        // so split it and find the correct answer
        var answers = s.split(/,\s*/);
        var correctChoice = -1;
        for (var i = 0; i < answers.length; ++i) {
            var ans = answers[i];

            // look for points...
            if (ans.search(/^_points:\s*(\d+)/i) > -1) {
                this.points_possible = RegExp.$1;

                // skip to the next answer
                continue;

            // or query text
            } else if (ans.search(/^_query_text:\s*(.*)/i) > -1) {
                this.query_text = RegExp.$1;

                // skip to the next answer
                continue;


            // or metadata -- starts with "_[a-zA-Z]"
            } else if (ans.search(/^\_[a-zA-Z]/) > -1) {
                this.extractMetadata(ans);

                // skip to the next answer
                continue;

            // if we've found the correct answer...
            } else if (ans.substr(0,1) == "*") {
                // if we already got a correctChoice, there's an error
                if (this.correct_answer_index != null) {
                    return "<b>error in checkbox answer (too many correct answers): " + s + "</b>";
                }

                // otherwise this one is correct
                this.correct_answer_index = i;

                // strip the leading *
                ans = ans.substr(1);
            }

            // convert &#42;s to literal *'s now; also convert commas
            ans = ans.replace(/\&#42;/g, '*');
            ans = ans.replace(/\&#44;/g, ',');

            // transfer to choices, trimming in the process
            this.choices.push(jQuery.trim(ans));
        }

        // note: answers will be coded as 0 and 1, with correctChoice numbered accordingly
        // (e.g. if first answer is correct, correctChoice is 0)

        // if we didn't get a correct choice, or if there aren't exactly 2 items,
        // that's an error
        if (this.correct_answer_index == null) {
            return "<b>error in checkbox answer (no correct answer specified): " + s + "</b>";
        } else if (this.choices.length != 2) {
            return "<b>error in checkbox answer (more or less than 2 choices specified): " + s + "</b>";
        }

        //safe_log(this);
    },


    getHTML: function(mode) {

        //safe_log("CB.getHTML()");
        //safe_log(this);

	var html = "";

	// CB's never have query text

	// if we're in preview mode or
	// (a student answer has been given and we're not in review_correct_incorrect mode),
	// or the student got it correct
	// disable the checkbox
	var disabled = "";
	if (mode == 'preview' || (mode != 'review_correct_incorrect' && this.user_answer_index != null) || this.isCorrect()) {
	    disabled = "disabled";
	}

	// if user has answered with choice 0, it should be checked, unless we're in preview mode
	var checked = "";
	if (this.user_answer_index === 0 && mode != 'preview') {
	    checked = "checked"
	}

        /*
        if (mode === "review") {
            var span_class = "visually-hidden query_cb answered";
            var checkStatus = "Not Checked ";
            if (this.user_answer_index === 0) {
                span_class += " checked";
                checkStatus = "Checked ";
            }

            if (this.label) {
                html += '<span class="cb_label ' + checked + '">' + this.label + '</span>';
            }

            html += '<span type="checkbox" class="' + span_class  + '" id="query_answer_' + this.query_index + '">(';
            html += checkStatus;
            html += ")</span>";
            this.removeLabel();
        }
        */
        //else {
	       //html += '<input aria-nodescribedby="' + this.parent_question.md.legend + '" type="checkbox" class="query_cb" id="query_answer_' + this.query_index + '" ' + disabled + ' ' + checked + '>';
            html += '<input type="checkbox" class="query_cb" id="query_answer_' + this.query_index + '" ' + disabled + ' ' + checked + '>';
           if (this.label) {
                html += '<label for="query_answer_' + this.query_index + '">' + this.label + '</label>';
           }
	       html += this.finish_inline_query(this.choices[this.correct_answer_index], mode);
        //}

	return html;
    }

});

/* Google's inert polyfill */
/* https://cdn.rawgit.com/GoogleChrome/inert-polyfill/v0.1.0/inert-polyfill.min.js */
window.addEventListener("load", function () {
    function h(a, b, c) { if (0 > b) { if (a.previousElementSibling) { for (a = a.previousElementSibling; a.lastElementChild;)a = a.lastElementChild; return a } return a.parentElement } if (a != c && a.firstElementChild) return a.firstElementChild; for (; null != a;) { if (a.nextElementSibling) return a.nextElementSibling; a = a.parentElement } return null } function g(a) { for (; a && a !== document.documentElement;) { if (a.hasAttribute("inert")) return a; a = a.parentElement } return null } (function (a) {
        var b = document.createElement("style");
        b.type = "text/css"; b.styleSheet ? b.styleSheet.cssText = a : b.appendChild(document.createTextNode(a)); document.body.appendChild(b)
    })("/*[inert]*/[inert]{position:relative!important;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;pointer-events:none}[inert]::before{content:'';display:block;position:absolute;top:0;left:0;right:0;bottom:0}"); var c = 0; document.addEventListener("keydown", function (a) { c = 9 === a.keyCode ? a.shiftKey ? -1 : 1 : 0 }); document.addEventListener("mousedown",
        function () { c = 0 }); document.body.addEventListener("focus", function (a) {
            var b = a.target, f = g(b); if (f) {
                if (document.hasFocus() && 0 !== c) {
                    var d = document.activeElement, e = new KeyboardEvent("keydown", { keyCode: 9, which: 9, key: "Tab", code: "Tab", keyIdentifier: "U+0009", shiftKey: !!(0 > c), bubbles: !0 }); Object.defineProperty(e, "keyCode", { value: 9 }); document.activeElement.dispatchEvent(e); if (d != document.activeElement) return; for (d = f; ;) {
                        d = h(d, c, f); if (!d) break; a: {
                            e = b; if (!(0 > d.tabIndex) && (d.focus(), document.activeElement !== e)) {
                                e =
                                !0; break a
                            } e = !1
                        } if (e) return
                    }
                } b.blur(); a.preventDefault(); a.stopPropagation()
            }
        }, !0); document.addEventListener("click", function (a) { g(a.target) && (a.preventDefault(), a.stopPropagation()) }, !0)
});

var Standard_Dialog = function () {
    var $manuscript = $('#manuscript');
    var $body = $('body');
    var $dialog, $content;
    var options = {};

    return {
        open: function (html, opt) {
            // Extend default options
            options = $.extend(true, (opt || {}), {
                width: 450,
                modal: true,
                draggable: false,
                resizable: false
            });

            // Set buttons
            if (options.buttons === 'none') {
                options.buttons = null;
            }
            else if (options.buttons === 'OK' || !options.buttons) {
                options.buttons = [{ text: 'OK', click: Standard_Dialog.close }];
            }


            // Remove previous dialog
            $('#standard_dialog_div').dialog('close').remove();

            if (options.modal) {
                $body.css('overflow', 'hidden');
            }

            // Hide other content from reader for mobile devices
            $manuscript.attr('aria-hidden', 'true');

            $dialog = $('<div id="standard_dialog_div"></div>');

            //$content = $('<div id="dialog_mesg" class="dialog-div">' + (html || '') + '</div>');
            $content = (html || '');

            var activeEl = document.activeElement;
            var fneScrollTop = undefined;
            if (activeEl.classList.contains('question_submit_button')) {

                var activeElPos = activeEl.getBoundingClientRect();

                console.log('active element positioning: ', activeElPos);

                try {
                    var fne_content = window.parent.document.getElementById('fne-content');
                    var fneScrollTop = fne_content.scrollTop;

                    console.log('fneScrollTop: ' + fneScrollTop);
                }
                catch (e) {
                    console.log('caught error trying to get fne-content: ' + e.message);
                    fneScrollTop = undefined;
                }
            }

            $dialog.append($content);
            $dialog.dialog(options);
            
            if (fneScrollTop !== undefined) {
                var popUpTop;
                if (fneScrollTop < activeElPos.top) {
                    popUpTop = fneScrollTop + ((activeElPos.top - fneScrollTop) / 2)
                }
                else {
                    popUpTop = fneScrollTop;
                }
                $('.ui-dialog.ui-widget').css({ top: popUpTop + 'px' });
            }


            // If we open toc we use as description only instruction in the top of dialog and not links
            //var isMenu = $content.find('#toc_head').length;
            //var labelledBy = isMenu ? 'toc_head' : 'standard_dialog_div';

            var diag_title = '';
            var labelledby = 'standard_dialog_div';
            if (typeof opt.diagTitle === 'string') {
                diag_title = '<h1 id="diag-title">' + opt.diagTitle + '</h1>';
                $('#standard_dialog_div').before(diag_title);
                labelledby = 'diag-title';
            }

            var $parentDialog = $dialog.parent().attr({
                'tabindex': '-1',
                'role': 'dialog',
                //'aria-describedby': 'standard_dialog_div',
                'aria-labelledby': labelledby,
                'aria-model': 'true'
            });

            /*
            $('#standard_dialog_div').appendTo($parentDialog);
            $parentDialog.css('display', 'flex');
            */

            /*
            $parentDialog.find('#ui-dialog-title-standard_dialog_div').attr({
                'role': 'heading',
                'aria-level': '1'
            });
            */

            // Remove close btn
            $('.ui-dialog-titlebar-close').remove();

            var $lastBtn = $parentDialog.find('button').last();
            var $firstBtn = $parentDialog.find('button').first();

            // Attach events
            $parentDialog.bind('keydown', function (e) {
                var code = e.keyCode || e.which;

                if (code === KEY_CODES.ESC) {
                    if (typeof options.onEsc === 'function') {
                        options.onEsc();
                    }
                    else if (typeof options.buttons[0].click === 'function') {
                        options.buttons[0].click();
                    }
                    else {
                        Standard_Dialog.close(true);
                    }
                    //Standard_Dialog.close(true);
                }

                if (e.target === this && e.shiftKey && code === KEY_CODES.TAB) {
                    e.preventDefault();
                    $lastBtn.focus();
                }
            });
            $lastBtn.bind('keydown', function (e) {
                var code = e.keyCode || e.which;

                if (!e.shiftKey && code === KEY_CODES.TAB) {
                    $firstBtn.focus();
                    e.preventDefault();
                }
            });
            $firstBtn.bind('keydown', function (e) {
                var code = e.keyCode || e.which;

                if (e.shiftKey && code === KEY_CODES.TAB) {
                    if ($lastBtn.length > 0) {
                        $lastBtn.focus();
                    }
                    e.preventDefault();
                }
            });

            //if (Platform.mobile) {
                // For android we set focus on dialog in any cases
            //setTimeout(function () {
                //$firstBtn.focus();
                $parentDialog.focus();
                //$('#diag-title').focus();
            //}, 100);
            //}
            /*
            else if (isMenu) {
                $parentDialog.find('a').first().focus();
            }
            else {
                if (options.buttons.length === 2) {
                    $lastBtn.focus();
                }
                else {
                    $firstBtn.focus();
                }
            }
            */
        },

        // for now alert is just like any other dialog; we could change that later.
        wait: function (html, opt) {
            options = opt || {};
            options.buttons = 'none';
            Standard_Dialog.open(html, options);
        },

        alert: function (html, opt) {
            Standard_Dialog.open(html, opt);
        },

        close: function () {
            // Show other content
            $manuscript.removeAttr('aria-hidden');

            if ($dialog) {
                $dialog.dialog('close').remove();
            }
            else {
                $("#standard_dialog_div").remove();
            }

            console.log(options);

            if (options.from) {
                options.from.focus();
            }

            if (options.modal) {
                $body.css('overflow', 'auto');
            }
        }
    };
}();

var KEY_CODES = {
    SPACE: 32,
    ENTER: 13,
    ESC: 27,
    TAB: 9
};

Metadata = Metadata.extend({

    add_from_jq: function (jq) {
        if (jq == null || jq.length == 0) {
            return;
        }

        // copy out the html so we can paste it back in at the end
        var original_html = jq.html();

        // first extract <field> key-value pairs
        var md = this;

        jq.find("[data-type=field]").each(function (index, element) {
            var key = $(element).attr("data-key");
            var value = $(element).html();
            md.add(key, value);
        });


        // then remove these
        jq.find("[data-type=field]").remove();

        // and look for "old-style" lines (_key: value)
        var lines = jq.html().split("\n");
        for (var i = 0; i < lines.length; ++i) {
            var l = $.trim(lines[i]).replace(/^_/, "");
            if (l == "") continue;

            if (l.search(/^(.*?)\s*:\s*(.*)/) > -1) {
                md.add(RegExp.$1, RegExp.$2);
            }
            else {
                safe_log("bad metadata line: " + l);
            }
        }

        // now restore the original html
        jq.html(original_html);
    }
});

function viewingInPX() {
    if (typeof window.PX !== 'undefined') {
        return true;
    }
    if (typeof window.parent.PX !== 'undefined') {
        return true;
    }
    return false;
}
